# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_mscolab
    ~~~~~~~~~~~~~~~~~~~

    This module is used to test mscolab related gui.

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
from mslib.msui.mss_qt import QtWidgets, QtTest, QtCore
import logging
import time

from mslib.mscolab.server import db, app, initialize_managers
from mslib._tests.constants import MSCOLAB_URL_TEST
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import Project
import mslib.msui.mscolab as mc


class Test_Mscolab(object):
    def setup(self):
        logging.debug("starting")
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = mc.MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                          mscolab_server_url=MSCOLAB_URL_TEST)
        self.window.show()
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()

        self.app = app
        self.app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
        self.app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
        self.app, _, cm, fm = initialize_managers(self.app)
        self.fm = fm
        self.cm = cm
        db.init_app(self.app)

    def teardown(self):
        # to disconnect connections, and clear token
        self.window.logout()
        for window in self.window.active_windows:
            window.hide()
        self.window.hide()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    def test_url_combo(self):
        assert self.window.url.count() >= 1

    def test_login(self):
        self._login()
        # screen shows logout button
        assert self.window.loggedInWidget.isVisible() is True
        assert self.window.loginWidget.isVisible() is False
        # test project listing visibility
        assert self.window.listProjects.model().rowCount() == 2
        # test logout
        QtTest.QTest.mouseClick(self.window.logoutButton, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert self.window.loggedInWidget.isVisible() is False
        assert self.window.loginWidget.isVisible() is True

    def test_disconnect(self):
        self._connect_to_mscolab()
        QtTest.QTest.mouseClick(self.window.disconnectMscolab, QtCore.Qt.LeftButton)
        assert self.window.mscolab_server_url is None

    def _connect_to_mscolab(self):
        self.window.url.setEditText("http://localhost:8084")
        QtTest.QTest.mouseClick(self.window.connectMscolab, QtCore.Qt.LeftButton)
        time.sleep(0.5)

    def _login(self):
        # login
        self._connect_to_mscolab()
        self.window.emailid.setText('a')
        self.window.password.setText('a')
        QtTest.QTest.mouseClick(self.window.loginButton, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()

    def test_activate_project(self):
        self._login()
        # activate a project
        self._activate_project_at_index(0)
        assert self.window.active_pid is not None

    def _activate_project_at_index(self, index):
        item = self.window.listProjects.item(index)
        point = self.window.listProjects.visualItemRect(item).center()
        QtTest.QTest.mouseClick(self.window.listProjects.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseDClick(self.window.listProjects.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()

    def test_view_open(self):
        self._login()
        # test without activating project
        QtTest.QTest.mouseClick(self.window.topview, QtCore.Qt.LeftButton)
        QtTest.QTest.mouseClick(self.window.sideview, QtCore.Qt.LeftButton)
        QtTest.QTest.mouseClick(self.window.tableview, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert len(self.window.active_windows) == 0
        # test after activating project
        self._activate_project_at_index(0)
        QtTest.QTest.mouseClick(self.window.tableview, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert len(self.window.active_windows) == 1
        QtTest.QTest.mouseClick(self.window.topview, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert len(self.window.active_windows) == 2
        QtTest.QTest.mouseClick(self.window.sideview, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        assert len(self.window.active_windows) == 3

    def test_save_fetch(self):

        self._login()
        self._activate_project_at_index(0)
        # change waypoint
        self.window.waypoints_model.invert_direction()
        wpdata1 = self.window.waypoints_model.waypoint_data(0)
        # don't save, fetch
        QtTest.QTest.mouseClick(self.window.fetch_ft, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        wpdata2 = self.window.waypoints_model.waypoint_data(0)
        assert wpdata2.lat != wpdata1.lat
        # save, fetch
        self.window.waypoints_model.invert_direction()
        wpdata1 = self.window.waypoints_model.waypoint_data(0)
        QtTest.QTest.mouseClick(self.window.save_ft, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        time.sleep(2)  # for change to take place
        QtTest.QTest.mouseClick(self.window.fetch_ft, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        wpdata2 = self.window.waypoints_model.waypoint_data(0)
        assert wpdata2.lat == wpdata1.lat
        # to undo changes
        self.window.waypoints_model.invert_direction()
        QtTest.QTest.mouseClick(self.window.save_ft, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        # to let server process
        time.sleep(2)

    def test_autosave(self):
        self._login()
        self._activate_project_at_index(0)
        QtTest.QTest.mouseClick(self.window.autoSave, QtCore.Qt.LeftButton,
                                pos=QtCore.QPoint(2, self.window.autoSave.height() / 2))
        QtWidgets.QApplication.processEvents()
        # sleeping to let server do the change
        time.sleep(3)
        with self.app.app_context():
            project = Project.query.filter_by(id=self.window.active_pid).first()
        assert project.autosave is self.window.autoSave.isChecked()
        QtTest.QTest.mouseClick(self.window.autoSave, QtCore.Qt.LeftButton,
                                pos=QtCore.QPoint(2, self.window.autoSave.height() / 2))
        QtWidgets.QApplication.processEvents()
        # top let server process
        time.sleep(2)



# -*- coding: utf-8 -*-
"""

    mslib.mscolab._tests.test_chat
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    tests for chat functionalities

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import socketio
import requests
import json
import datetime

from mslib.mscolab.models import Message
from mslib._tests.constants import MSCOLAB_URL_TEST
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.server import db, app, initialize_managers


class Test_Chat(object):

    def setup(self):
        self.sockets = []
        self.app = app
        self.app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
        self.app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
        self.app, _, cm, _ = initialize_managers(self.app)
        self.cm = cm
        db.init_app(self.app)

    def test_send_message(self):
        r = requests.post(MSCOLAB_URL_TEST + "/token", data={
                          'email': 'a',
                          'password': 'a'
                          })
        response = json.loads(r.text)
        # standard Python
        sio = socketio.Client()

        def handle_chat_message(message):
            pass

        sio.on('chat-message-client', handler=handle_chat_message)
        sio.connect(MSCOLAB_URL_TEST)
        sio.emit('start', response)
        sio.sleep(2)
        self.sockets.append(sio)
        sio.emit("chat-message", {
                 "p_id": 1,
                 "token": response['token'],
                 "message_text": "message from 1"
                 })
        # testing non-ascii message
        sio.emit("chat-message", {
                 "p_id": 1,
                 "token": response['token'],
                 "message_text": "® non ascii"
                 })
        sio.sleep(2)
        with self.app.app_context():
            message = Message.query.filter_by(text="message from 1").first()
            assert message.p_id == 1
            assert message.u_id == 8

            message = Message.query.filter_by(text="® non ascii").first()
            assert message is not None
            Message.query.filter_by(text="message from 1").delete()
            Message.query.filter_by(text="® non ascii").delete()
            db.session.commit()

    def test_get_messages(self):
        r = requests.post(MSCOLAB_URL_TEST + "/token", data={
                          'email': 'a',
                          'password': 'a'
                          })
        response = json.loads(r.text)
        # standard Python
        sio = socketio.Client()

        def handle_chat_message(message):
            pass

        sio.on('chat-message-client', handler=handle_chat_message)
        sio.connect(MSCOLAB_URL_TEST)
        sio.emit('start', response)
        sio.sleep(2)
        self.sockets.append(sio)
        sio.emit("chat-message", {
                 "p_id": 1,
                 "token": response['token'],
                 "message_text": "message from 1"
                 })
        sio.emit("chat-message", {
                 "p_id": 1,
                 "token": response['token'],
                 "message_text": "message from 1"
                 })
        sio.sleep(4)
        with self.app.app_context():
            messages = self.cm.get_messages(1)
            assert len(messages) == 2
            assert messages[0]["user"] == 8

            messages = self.cm.get_messages(1, last_timestamp=datetime.datetime(1970, 1, 1))
            assert len(messages) == 2
            assert messages[0]["user"] == 8

            messages = self.cm.get_messages(1, last_timestamp=datetime.datetime.now())
            assert len(messages) == 0

    def test_get_messages_api(self):
        r = requests.post(MSCOLAB_URL_TEST + "/token", data={
                          'email': 'a',
                          'password': 'a'
                          })
        response = json.loads(r.text)
        token = response["token"]
        data = {
            "token": token,
            "p_id": 1,
            "timestamp": datetime.datetime(1970, 1, 1).strftime("%m %d %Y, %H:%M:%S")
        }
        # returns an array of messages
        r = requests.post(MSCOLAB_URL_TEST + "/messages", data=data)
        response = json.loads(r.text)
        assert len(response["messages"]) == 2

        data["token"] = "dummy"
        # returns False due to bad authorization
        r = requests.post(MSCOLAB_URL_TEST + "/messages", data=data)
        assert r.text == "False"
        with self.app.app_context():
            Message.query.filter_by(text="message from 1").delete()
            db.session.commit()

    def teardown(self):
        for socket in self.sockets:
            socket.disconnect()



# -*- coding: utf-8 -*-
"""

    mslib.msui.lagranto_output_reader
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    A class that reads all output data files generated by Lagranto in
    a specified directory.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import sys

import copy
import logging
import os
import pickle
import numpy
from fs import open_fs

import warnings
warnings.warn("the lagranto_output_reader module is deprecated", DeprecationWarning,
              stacklevel=2)


class LagrantoOutputReader(object):
    """A class that reads all output data files generated by Lagranto in
       a specified directory.

       Usage:
======

    lout = LagrantoOutputReader(<complete path to output files>)

That's it! After this call the data of trajectories read from lsl* files
can be accessed via

    lout.data[trajectoryIndex][<variable_name>],

the data of lsl* files (statistics of trajectories in ls* files) via

    lout.stats[<output_filename>][<variable_name>],

where <trajectoryIndex> is the number of the trajectory (increasing with each
file) (if more than one trajectories were computed from a start region or in
scf mode; if there is only one file with one trajectory, this number takes
the value 0).

Trajectory data is stored in NumPy arrays.

Call

    len(lout.data)

to see how many trajectories have been read, and
which files have been read, and

    lout.data[trajectoryIndex].keys()

to see which variables are available for a trajectory. Call

    lout.meta[trajectoryIndex]

to view the metadata available for a given trajectory.


Example:
========

    import lagrantooutputreader
    lout = lagrantooutputreader.LagrantoOutputReader( \
                'output/ntr_20070401_00_f24_1_j_o30/')
    len(lout.data)
    lout.data[10].keys()
    lout.data[10]['PV']
    lout.meta[10]
    .. etc ..
    """

    def __init__(self, lagrantoOutputPath, of=sys.stdout):
        """Initialise the reader and read all ls* and lsl* files contained
           in the path argument.

        Arguments:
        lagrantoOutputPath -- path to the directory containing the Lagranto
                              output files. All lsl* and ls* files in this
                              directory will be read.

        Keyword arguments:
        of -- output for all print statements, has to be an object
              with a *write()* method, e.g. a file; if not
              specified sys.stdout is taken
        """

        #
        # Expand path to user home if necessary and test if the argument is
        # a valid path. Raise an exception if not.
        self.lagrantoOutputPath = os.path.expanduser(lagrantoOutputPath)
        if not os.path.isdir(self.lagrantoOutputPath):
            raise ValueError("Argument " +
                             self.lagrantoOutputPath +
                             " does not represent a valid path.")

        self.of = of
        self.data = []  # stores trajectory data from lsl* files
        self.meta = []  # stores metadata belonging to the trajectories
        self.stats = {}  # stores data from ls* files

        #
        # Read all files contained in lagrantoOutputPath.
        self.__readFilesInDirectory()

    def __read_lslFile(self, fname):
        """Parse a Lagranto lsl-file and reads the data into NumPy arrays.
           Also read the corresponding metadata files, if existent.

        Data contained in the file is accessible under
            self.data[trajectoryIndex][variablename]
            self.meta[trajectoryIndex][metatagname]
        after this method has been executed.

        self.meta[trajectoryIndex] will contain at least the variables 'file'
        and 'startttime_filename', indicating the file from which the trajectory
        was read and the start time, determined from the filename or the
        directory name (datetime object). If a metadata file is present,
        the fields 'startcoordinates', 'startttime' and 'duration' will contain
        values from the start points list passed to LagrantoWrapper,
        metadata additionally specified together with the start coordinates
        will also be added to self.meta[trajectoryIndex].

        Arguments:
        fname -- file name of the lsl-file, without path (i.e. only 'lsl_...').
        """

        logging.debug("Parsing output file '%s'", fname)

        #
        # Open the lsl text file and read its contents into the list 'lines',
        # containing the individual text lines as elements.
        _dirname, _name = os.path.split(os.path.join(self.lagrantoOutputPath, fname))
        _fs = open_fs(_dirname)
        flsl = _fs.open(_name, 'r')
        lines = flsl.readlines()

        #
        # Open and read the metadata file, if existent.
        metafile = os.path.join(self.lagrantoOutputPath, fname) + ".meta.pyl"
        _dirname, _name = os.path.split(metafile)
        _fs = open_fs(_dirname)
        if _fs.exists(_name):
            logging.debug("Reading metadata file '%s'", os.path.basename(metafile))
            # ToDo example file without UnicodeDecodeError: 'ascii' codec can't decode byte 0xfa in
            # position 1: ordinal not in range(128)
            with _fs.open(_name, "rb") as fileobj:
                startCoordinates = pickle.load(fileobj)
                startTime = pickle.load(fileobj)
                duration = pickle.load(fileobj)

        else:
            logging.debug("No metadata file '%s' found.", metafile)
            startCoordinates = None
            startTime = None
            duration = None

        #
        # Data contained in the current file will be appended to self.data,
        # hence be accessible under self.data[startIndex + X], where
        # startIndex is the length of self.data at the time this method is
        # called.
        startIndex = len(self.data)
        #
        # Get the current filename (without path) and determine the trajectory
        # start time from this name. These two variable will be stored in
        # self.meta[trajectoryIndex]['file'] and ['starttime_filename'],
        # respectively. Even if no metadata file was found, these two
        # variables will be available for each trajectory.
        filename = os.path.basename(fname)
        startTimeFile = self.__getDateTimeFromName(filename)
        if not startTimeFile:
            startTimeFile = self.initTimeDir

        #
        # Start to parse the file: Get the index of the first blank line in
        # order to skip the header.
        firstBlank = lines.index(" \n")
        #
        # Variable names are contained in the line following the first
        # blank line. Initialise empty lists for each variable.
        varNames = lines[firstBlank + 1].split()
        emptyVariableDictionaryTemplate = {}
        for varName in varNames:
            emptyVariableDictionaryTemplate[varName] = []
        self.data.append(copy.deepcopy(emptyVariableDictionaryTemplate))

        #
        # Iterate through the lines containing the data (lines 4ff. after
        # the first blank line. Append the numerical values, converted to
        # floats, to the data lists identified by the variable names (the
        # [-1] index references the last element). If a blank line is
        # encountered, this indicates a new trajectory in the file. Hence,
        # append a new trajectory to self.data.
        for dataLine in lines[firstBlank + 4:]:
            if dataLine == " \n":
                self.data.append(
                    copy.deepcopy(emptyVariableDictionaryTemplate))
            for varName, strValue in zip(varNames, dataLine.split()):
                self.data[-1][varName].append(float(strValue))

        #
        # Finally, for each trajectory from this file: convert the lists
        # with the numerical data to NumPy arrays for faster access; create
        # a metadata dictionary (containing at least the filename and start
        # time determined from this name or the directory name) and append
        # it to self.meta.
        for i in range(startIndex, len(self.data)):
            for varName in varNames:
                self.data[i][varName] = numpy.array(self.data[i][varName])
            metadict = {'file': filename,
                        'starttime_filename': startTimeFile}
            # If metadata is present, check if a metadata dictionary was
            # specified for this trajectory (last element of the start
            # coordinates list is a dictionary, cf. lagranto.py). If yes,
            # add the contained elements to self.meta. Store start coordinates,
            # time and duration in the metadata dictionary.
            if startCoordinates is not None:
                if isinstance(startCoordinates[i - startIndex][-1], dict):
                    metadict.update(startCoordinates[i - startIndex].pop())
                metadict['startcoordinates'] = startCoordinates[i - startIndex]
                metadict['starttime'] = startTime
                metadict['duration'] = duration
            self.meta.append(metadict)

    def __read_lsFile(self, fname):
        """Parse a Lagranto ls-file and reads the data into NumPy arrays.

        Data contained in the file is accessible under
            self.stats[filename][variablename]
        after this method has been executed.

        Arguments:
        fname -- file name of the lsl-file, without path (i.e. only 'ls_...').
        """
        #
        # This method is similar to __read_lslFile(). Only those parts of the
        # source code are documented that differ from __read_lslFile().
        logging.debug("Parsing output file '%s'", fname)
        _dirname, _name = os.path.split(os.path.join(self.lagrantoOutputPath, fname))
        _fs = open_fs(_dirname)

        flsl = _fs.open(_name, 'r')
        lines = flsl.readlines()

        dataKey = os.path.basename(fname)
        self.stats[dataKey] = {}
        self.starttimes[dataKey] = self.__getDateTimeFromName(dataKey)
        if not self.starttimes[dataKey]:
            self.starttimes[dataKey] = self.initTimeDir

        firstBlank = lines.index(" \n")

        varNames = lines[firstBlank + 1].split()

        #
        # In the ls-files, all variables except for time, lat and lon are
        # stored with an additional column representing the standard deviation
        # of the trajectories for Lagranto runs in which more than one
        # trajectories have been computed. These columns aren't named
        # in the text files, hence we have to add their names to the
        # varNames list here. The convention used is the original variable
        # name appended by a '_stddev'. Note that the first three variabe
        # names (i.e. time, lat, lon) are skipped in this loop.
        for varName in varNames[3:]:
            varNames.insert(varNames.index(varName) + 1, varName + "_stddev")

        #
        # Also, an additional column following the variable data is stored in
        # the text file, containing the number of trajectories that have
        # left the model domain after time t. We append a new variable to our
        # list that will accomodate this information.
        varNames.append("evolution_of_ntraout")

        for varName in varNames:
            self.stats[dataKey][varName] = []

        #
        # Get the index of the second blank line to get the end of the data
        # section.
        secondBlank = lines.index(" \n", firstBlank + 1)

        #
        # Process the data section.
        for dataLine in lines[firstBlank + 3:secondBlank]:
            for varName, strValue in zip(varNames[:-1], dataLine.split()):
                self.stats[dataKey][varName].append(float(strValue))

        #
        # Process the second data section containing 'evolution_of_ntraout'.
        for dataLine in lines[secondBlank + 4:]:
            strValue = dataLine.split()[1]
            self.stats[dataKey]["evolution_of_ntraout"].append(float(strValue))

        for varName in varNames:
            self.stats[dataKey][varName] = numpy.array(self.stats[dataKey][varName])

    def __readFilesInDirectory(self):
        """Read all files contained in the directory passed to the constructor.

        Arguments:
        -- None --
        """
        #
        # Remove trailing '/' if present, so that os.path.split() returns the
        # last subdirectory (var[-1] is short for var[len(var)-1]).
        subdirname = self.lagrantoOutputPath
        if subdirname[-1] == '/':
            subdirname = subdirname[:-1]
        subdirname = os.path.split(subdirname)[1]

        #
        # Try to extract the init time from the directory name.
        self.initTimeDir = self.__getDateTimeFromName(subdirname)

        #
        # Get a list of files in the directory given by lagrantoOutputPath.
        fileList = os.listdir(self.lagrantoOutputPath)

        #
        # Extract all lsl* data files from the list and call the corresponding
        # read method.
        lslFiles = [fname for fname in fileList
                    if fname.startswith("lsl_") and
                    not fname.endswith(".meta.pyl")]
        for fname in lslFiles:
            self.__read_lslFile(fname)

        #
        # The same for all ls* data files.
        lsFiles = [fname for fname in fileList
                   if fname.startswith("ls_") and
                   not fname.endswith(".meta.pyl")]
        for fname in lsFiles:
            self.__read_lsFile(fname)

    def __getDateTimeFromName(self, name):
        """Look for a sequence of format YYYYMMDD_hh[mm] in name (where mm is)
           optional and return a datetime object.
        """
        #
        # Try to extract a time of format YYYYMMDD_hh[mm] from 'name'.
        # Start at each index from the beginning of the string 'name'
        # to len(name)-11[13] (the time is 11[13] characters long) and try
        # to convert the part of the string to a datetime object. If successful,
        # break out of the loop. If the format YYYYMMDD_hhmm is successful,
        # return this value, otherwise try YYYYMMDD_hh.
        timeFormat = ["%Y%m%d_%H%M", "%Y%m%d_%H"]  # YYYYMMDD_hh[mm]
        timeFormatLen = [13, 11]

        for tf, tflen in zip(timeFormat, timeFormatLen):
            for i in range(len(name) - tflen):
                try:
                    time = datetime.datetime.strptime(name[i:i + tflen], tf)
                except ValueError as er:
                    logging.error("%s %s", er, type(er))
                else:
                    return time
        return None



# -*- coding: utf-8 -*-
"""

    mslib.mscolab.server
    ~~~~~~~~~~~~~~~~~~~~

    Server for mscolab module

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from flask import Flask, request, g
import logging
import json
import datetime
import functools
from validate_email import validate_email
import socketio

from mslib.mscolab.models import User, db, Change
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.sockets_manager import setup_managers
# set the project root directory as the static folder
app = Flask(__name__, static_url_path='')
app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
app.config['SECRET_KEY'] = mscolab_settings.SECRET_KEY

log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)


def initialize_managers(app):
    sockio, cm, fm = setup_managers(app)
    # initiatializing socketio and db
    app.wsgi_app = socketio.Middleware(socketio.server, app.wsgi_app)
    sockio.init_app(app)
    db.init_app(app)
    return (app, sockio, cm, fm)


_app, sockio, cm, fm = initialize_managers(app)


def check_login(emailid, password):
    user = User.query.filter_by(emailid=str(emailid)).first()
    if user:
        if user.verify_password(password):
            return user
    return False


def register_user(email, password, username):
    user = User(email, username, password)
    user_exists = User.query.filter_by(emailid=str(email)).first()
    is_valid = validate_email(email)
    if not is_valid:
        return 'False'
    if user_exists:
        return 'False'
    user_exists = User.query.filter_by(username=str(username)).first()
    if user_exists:
        return 'False'
    db.session.add(user)
    db.session.commit()
    return 'True'


def verify_user(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        user = User.verify_auth_token(request.form.get('token', False))
        if not user:
            return "False"
        else:
            # saving user details in flask.g
            g.user = user
            return func(*args, **kwargs)
    return wrapper


# ToDo setup codes in return statements
@app.route("/")
def hello():
    return "Mscolab server"


# User related routes


@app.route('/token', methods=["POST"])
def get_auth_token():
    emailid = request.form['email']
    password = request.form['password']
    user = check_login(emailid, password)
    if user:
        token = user.generate_auth_token()
        return json.dumps({
                          'token': token.decode('ascii'),
                          'user': {'username': user.username, 'id': user.id}})
    else:
        logging.debug("Unauthorized user: %s", emailid)
        return "False"


@app.route('/test_authorized')
def authorized():
    token = request.values['token']
    user = User.verify_auth_token(token)
    if user:
        return "True"
    else:
        return "False"


@app.route("/register", methods=["POST"])
def user_register_handler():
    email = request.form['email']
    password = request.form['password']
    username = request.form['username']
    return register_user(email, password, username)


@app.route('/user', methods=["GET"])
@verify_user
def get_user():
    return json.dumps({'user': {'id': g.user.id, 'username': g.user.username}})


# Chat related routes
@app.route("/messages", methods=['POST'])
@verify_user
def messages():
    timestamp = datetime.datetime.strptime(request.form['timestamp'], '%m %d %Y, %H:%M:%S')
    p_id = request.form.get('p_id', None)
    messages = cm.get_messages(p_id, last_timestamp=timestamp)
    return json.dumps({'messages': messages})


# File related routes
@app.route('/create_project', methods=["POST"])
@verify_user
def create_project():
    path = request.values['path']
    description = request.values['description']
    content = request.values.get('content', None)
    user = g.user
    return str(fm.create_project(path, description, user, content=content))


@app.route('/get_project', methods=['GET'])
@verify_user
def get_project():
    p_id = request.values.get('p_id', None)
    user = g.user
    result = fm.get_file(int(p_id), user)
    if result is False:
        return "False"
    return json.dumps({"content": result})


@app.route('/get_changes', methods=['GET'])
@verify_user
def get_changes():
    p_id = request.values.get('p_id', None)
    user = g.user
    result = fm.get_changes(int(p_id), user)
    if result is False:
        return "False"
    return json.dumps({"changes": result})


@app.route('/get_change_id', methods=['GET'])
@verify_user
def get_change_by_id():
    ch_id = request.values.get('ch_id', None)
    user = g.user
    result = fm.get_change_by_id(int(ch_id), user)
    if result is False:
        return "False"
    return json.dumps({"change": result})


@app.route('/authorized_users', methods=['GET'])
@verify_user
def authorized_users():
    p_id = request.values.get('p_id', None)
    return json.dumps({"users": fm.get_authorized_users(int(p_id))})


@app.route('/projects', methods=['GET'])
@verify_user
def get_projects():
    user = g.user
    return json.dumps({"projects": fm.list_projects(user)})


@app.route('/delete_project', methods=["POST"])
@verify_user
def delete_project():
    p_id = request.form.get('p_id', None)
    user = g.user
    return str(fm.delete_file(int(p_id), user))


@app.route('/add_permission', methods=['POST'])
@verify_user
def add_permission():
    p_id = request.form.get('p_id', 0)
    u_id = request.form.get('u_id', 0)
    username = request.form.get('username', None)
    access_level = request.form.get('access_level', None)
    user = g.user
    if u_id == 0:
        user_v = User.query.filter_by(username=username).first()
        if user_v is None:
            return "False"
        u_id = user_v.id
    success = str(fm.add_permission(int(p_id), int(u_id), username, access_level, user))
    if success == "True":
        sockio.sm.join_collaborator_to_room(int(u_id), int(p_id))
        sockio.sm.emit_new_permission(int(u_id), int(p_id))
    return success


@app.route('/revoke_permission', methods=['POST'])
@verify_user
def revoke_permission():
    p_id = request.form.get('p_id', 0)
    u_id = request.form.get('u_id', 0)
    username = request.form.get('username', None)
    user = g.user
    if u_id == 0:
        user_v = User.query.filter_by(username=username).first()
        if user_v is None:
            return "False"
        u_id = user_v.id
    success = str(fm.revoke_permission(int(p_id), int(u_id), username, user))
    if success:
        sockio.sm.emit_revoke_permission(int(u_id), int(p_id))
        sockio.sm.remove_collaborator_from_room(int(u_id), int(p_id))
    return success


@app.route('/modify_permission', methods=['POST'])
@verify_user
def modify_permission():
    p_id = request.form.get('p_id', 0)
    u_id = request.form.get('u_id', 0)
    username = request.form.get('username', None)
    access_level = request.form.get('access_level', None)
    user = g.user
    if u_id == 0:
        user_v = User.query.filter_by(username=username).first()
        if user_v is None:
            return "False"
        u_id = user_v.id
    success = str(fm.update_access_level(int(p_id), int(u_id), username, access_level, user))
    if success == "True":
        sockio.sm.emit_update_permission(u_id, p_id)
    return success


@app.route('/update_project', methods=['POST'])
@verify_user
def update_project():
    p_id = request.form.get('p_id', None)
    attribute = request.form['attribute']
    value = request.form['value']
    user = g.user
    return str(fm.update_project(int(p_id), attribute, value, user))


@app.route('/project_details', methods=["GET"])
@verify_user
def get_project_details():
    p_id = request.form.get('p_id', None)
    user = g.user
    return json.dumps(fm.get_project_details(int(p_id), user))


@app.route('/undo', methods=["POST"])
@verify_user
def undo_ftml():
    ch_id = request.form.get('ch_id', -1)
    ch_id = int(ch_id)
    user = g.user
    result = fm.undo(ch_id, user)
    # get p_id from change
    ch = Change.query.filter_by(id=ch_id).first()
    if result is True:
        sockio.sm.emit_file_change(ch.p_id)
    return str(result)


def start_server(app, sockio, cm, fm, port=8083):
    sockio.run(app, port=port)


def main():
    from mslib.mscolab.demodata import create_data
    # create data if not created
    create_data()
    start_server(_app, sockio, cm, fm)


# for wsgi
application = socketio.WSGIApp(sockio)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mslib/msui/ui/ui_mscolab_project_window.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MscolabProject(object):
    def setupUi(self, MscolabProject):
        MscolabProject.setObjectName("MscolabProject")
        MscolabProject.setWindowModality(QtCore.Qt.NonModal)
        MscolabProject.setEnabled(True)
        MscolabProject.setMinimumSize(QtCore.QSize(600, 400))
        self.centralwidget = QtWidgets.QWidget(MscolabProject)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setSizeConstraint(QtWidgets.QLayout.SetNoConstraint)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.username = QtWidgets.QLineEdit(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.username.sizePolicy().hasHeightForWidth())
        self.username.setSizePolicy(sizePolicy)
        self.username.setMinimumSize(QtCore.QSize(408, 0))
        self.username.setObjectName("username")
        self.horizontalLayout.addWidget(self.username)
        self.accessLevel = QtWidgets.QComboBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accessLevel.sizePolicy().hasHeightForWidth())
        self.accessLevel.setSizePolicy(sizePolicy)
        self.accessLevel.setObjectName("accessLevel")
        self.accessLevel.addItem("")
        self.accessLevel.addItem("")
        self.accessLevel.addItem("")
        self.horizontalLayout.addWidget(self.accessLevel)
        self.add = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.add.sizePolicy().hasHeightForWidth())
        self.add.setSizePolicy(sizePolicy)
        self.add.setObjectName("add")
        self.horizontalLayout.addWidget(self.add)
        self.modify = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.modify.sizePolicy().hasHeightForWidth())
        self.modify.setSizePolicy(sizePolicy)
        self.modify.setObjectName("modify")
        self.horizontalLayout.addWidget(self.modify)
        self.delete_1 = QtWidgets.QPushButton(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.delete_1.sizePolicy().hasHeightForWidth())
        self.delete_1.setSizePolicy(sizePolicy)
        self.delete_1.setObjectName("delete_1")
        self.horizontalLayout.addWidget(self.delete_1)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setSizeConstraint(QtWidgets.QLayout.SetMinimumSize)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.proj_info = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.proj_info.sizePolicy().hasHeightForWidth())
        self.proj_info.setSizePolicy(sizePolicy)
        self.proj_info.setText("")
        self.proj_info.setObjectName("proj_info")
        self.verticalLayout_4.addWidget(self.proj_info)
        self.user_info = QtWidgets.QLabel(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.user_info.sizePolicy().hasHeightForWidth())
        self.user_info.setSizePolicy(sizePolicy)
        self.user_info.setText("")
        self.user_info.setObjectName("user_info")
        self.verticalLayout_4.addWidget(self.user_info)
        self.verticalLayout_2.addLayout(self.verticalLayout_4)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.collaboratorsList = QtWidgets.QListWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.collaboratorsList.sizePolicy().hasHeightForWidth())
        self.collaboratorsList.setSizePolicy(sizePolicy)
        self.collaboratorsList.setMinimumSize(QtCore.QSize(256, 300))
        self.collaboratorsList.setObjectName("collaboratorsList")
        self.horizontalLayout_3.addWidget(self.collaboratorsList)
        self.messages = QtWidgets.QListWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.messages.sizePolicy().hasHeightForWidth())
        self.messages.setSizePolicy(sizePolicy)
        self.messages.setMinimumSize(QtCore.QSize(539, 300))
        self.messages.setObjectName("messages")
        self.horizontalLayout_3.addWidget(self.messages)
        self.verticalLayout_2.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.messageText = QtWidgets.QPlainTextEdit(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.messageText.sizePolicy().hasHeightForWidth())
        self.messageText.setSizePolicy(sizePolicy)
        self.messageText.setMinimumSize(QtCore.QSize(400, 50))
        self.messageText.setObjectName("messageText")
        self.horizontalLayout_2.addWidget(self.messageText)
        self.sendMessage = QtWidgets.QPushButton(self.centralwidget)
        self.sendMessage.setDefault(True)
        self.sendMessage.setObjectName("sendMessage")
        self.horizontalLayout_2.addWidget(self.sendMessage)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.changes = QtWidgets.QListWidget(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.changes.sizePolicy().hasHeightForWidth())
        self.changes.setSizePolicy(sizePolicy)
        self.changes.setMinimumSize(QtCore.QSize(400, 280))
        self.changes.setObjectName("changes")
        self.horizontalLayout_6.addWidget(self.changes)
        self.checkout = QtWidgets.QPushButton(self.centralwidget)
        self.checkout.setAutoDefault(False)
        self.checkout.setDefault(False)
        self.checkout.setObjectName("checkout")
        self.horizontalLayout_6.addWidget(self.checkout, 0, QtCore.Qt.AlignTop)
        self.verticalLayout_2.addLayout(self.horizontalLayout_6)
        MscolabProject.setCentralWidget(self.centralwidget)
        self.actionCloseWindow = QtWidgets.QAction(MscolabProject)
        self.actionCloseWindow.setObjectName("actionCloseWindow")
        MscolabProject.addAction(self.actionCloseWindow)

        self.retranslateUi(MscolabProject)
        self.actionCloseWindow.triggered.connect(MscolabProject.close)
        QtCore.QMetaObject.connectSlotsByName(MscolabProject)

    def retranslateUi(self, MscolabProject):
        _translate = QtCore.QCoreApplication.translate
        MscolabProject.setWindowTitle(_translate("MscolabProject", "Mscolab Project Chat"))
        self.accessLevel.setItemText(0, _translate("MscolabProject", "collaborator"))
        self.accessLevel.setItemText(1, _translate("MscolabProject", "admin"))
        self.accessLevel.setItemText(2, _translate("MscolabProject", "viewer"))
        self.add.setText(_translate("MscolabProject", "add"))
        self.modify.setText(_translate("MscolabProject", "modify"))
        self.delete_1.setText(_translate("MscolabProject", "delete"))
        self.messageText.setPlaceholderText(_translate("MscolabProject", "Enter message here"))
        self.sendMessage.setText(_translate("MscolabProject", "send"))
        self.checkout.setText(_translate("MscolabProject", "checkout"))
        self.actionCloseWindow.setText(_translate("MscolabProject", "CloseWindow"))
        self.actionCloseWindow.setShortcut(_translate("MscolabProject", "Ctrl+W"))




# -*- coding: utf-8 -*-
"""

    mslib.msui.trajectories_tool
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Tool application that manages trajectories.

    This file is part of mss.

    :copyright: Copyright 2008-2014 Deutsches Zentrum fuer Luft- und Raumfahrt e.V.
    :copyright: Copyright 2011-2014 Marc Rautenhaus (mr)
    :copyright: Copyright 2016-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
import os

from mslib.msui.mss_qt import QtCore, QtWidgets, QtGui, get_open_filename, get_existing_directory
from mslib.msui.mss_qt import ui_trajectories_window as ui
from mslib.msui import trajectory_item_tree as titree
from mslib.msui.viewwindows import MSSViewWindow
from mslib.msui.icons import icons
from mslib.msui import MissionSupportSystemDefaultConfig as mss_default
from mslib.utils import config_loader

import warnings
warnings.warn("the trajectories_tool module is deprecated", DeprecationWarning,
              stacklevel=2)


class MSSTrajectoriesToolWindow(MSSViewWindow, ui.Ui_TrajectoriesWindow):
    """Implements a trajectory tool.
    """

    name = "Trajectories Tool"
    moduleCloses = QtCore.pyqtSignal(name="moduleCloses")

    def __init__(self, parent=None, listviews=None, listtools=None, viewsChanged=None):
        """
        """
        super(MSSTrajectoriesToolWindow, self).__init__(parent)
        self.setupUi(self)
        self.last_save_directory = config_loader(dataset="data_dir", default=mss_default.data_dir)
        self.setWindowIcon(QtGui.QIcon(icons('64x64')))

        self.actionOpenFlightTrack.setEnabled(titree.HAVE_NAPPY)
        self.viewsChanged = viewsChanged

        # traj_item_tree stores all data corresponding to the map, i.e. flight
        # tracks, trajectories, etc. The design follows Qt's model/view
        # architecture: the data structure is derived from QAbstractItemModel,
        # thus it provides all methods necessary so that the data can be
        # displayed in a QTreeView widget, but it also makes the data accessible
        # for the rest of the program (especially the map).
        self.traj_item_tree = titree.LagrantoMapItemsTreeModel()

        # Connect the implementation for QtAbstractItemModel
        # (self.mapItemsTree) with the tree view. All changes in the data
        # structure will automatically be displayed in the GUI.
        self.tvVisibleElements.setModel(self.traj_item_tree)

        # List that stores the views the trajectory item tree is connected to.
        self.connected_views = []

        # Pointer to the QListWidget that accomodates the views that are
        # open in the MSUI.
        self.listviews = listviews
        self.listtools = listtools

        # Connect Qt SIGNALs:
        # ===================

        # Menu.
        self.actionOpenTrajectories.triggered.connect(self.loadTrajectories)
        self.actionOpenFlightTrack.triggered.connect(self.loadFlightTrack)

        # Item selection.
        self.btSelectMapElements.clicked.connect(self.selectMapElements)

        # Item style.
        self.btColour.clicked.connect(self.setCurrentItemColour)
        self.btLineStyle.clicked.connect(self.setCurrentItemLineStyle)
        self.btLineWidth.clicked.connect(self.setCurrentItemLineWidth)
        self.btTimeMarker.clicked.connect(self.setCurrentItemTimeMarker)

        # View control.
        self.btPlotInView.clicked.connect(self.plotCurrentItemInView)
        self.btRemoveFromView.clicked.connect(self.removeCurrentItemFromView)
        if self.viewsChanged is not None:
            self.viewsChanged.connect(self.updateViews)
        self.updateViews()

    def closeEvent(self, event):
        """Ask user if he/she wants to close the window.

        Overloads MSSViewWindow.closeEvent() and emits the signal
        'moduleCloses()'.
        """
        if super(MSSTrajectoriesToolWindow, self).closeEvent(event):
            self.moduleCloses.emit()

    def loadFlightTrack(self):
        """Slot for the 'Open Flight Track..' menu entry. Opens a file dialog
           and reads the selected NASA Ames file using nappy.
        """
        # Ask for a file to open, convert the return file name from type
        filename = get_open_filename(
            self, "Open NASA Ames File", "", "NASA Ames files (*.nas)",
            pickertag="filepicker_trajectories")
        # ToDo nappy needs filelike object first
        # _dirname, _name = os.path.split(filename)
        # _fs = open_fs(_dirname)
        if filename is not None:
            # until nappy knows filelike objects
            # nas_file = _fs.open(_name, "wb")
            nas_file = os.path.expanduser(filename)

            logging.debug("Loading flight track data from '%s'", nas_file)

            # Add the flight track to the data tree.
            try:
                new_item_index = self.traj_item_tree.addFlightTrack(nas_file)

                # Scroll the tree view so that the new item is visible and resize
                # the first column to be wide enough for the text.
                self.tvVisibleElements.scrollTo(new_item_index)
                self.tvVisibleElements.resizeColumnToContents(0)
            except RuntimeError:
                QtWidgets.QMessageBox.critical(
                    self, self.tr("Trajectory Tool"), self.tr("ERROR:\nCould not read file\n{}".format(nas_file)))

    def loadTrajectories(self):
        """Slot for the 'Open Trajectories..' menu entry. Opens a QFileDialog
           and reads the selected trajectory file using lagrantooutputreader.
        """
        # Ask for a directory to open.
        traj_dir = get_existing_directory(
            self, "Open Lagranto Output Directory", self.last_save_directory)
        if traj_dir is not None:
            logging.debug("Loading trajectory data from '%s'", traj_dir)

            # Test if selected directory contains subdirectories (as is
            # the case with ensemble runs). If yes,
            # load all those subdirectories. Otherwise load the selected
            # directory.
            # ToDo we have to solve this differently, it is not a good idea to have such a traversal
            traj_dir_items = os.listdir(traj_dir)
            subdirs = [sdir for sdir in traj_dir_items
                       if os.path.isdir(os.path.join(traj_dir, sdir))]
            if len(subdirs) > 0:
                for sdir in subdirs:
                    # Add the trajectories to the data tree.
                    new_item_index = self.traj_item_tree \
                        .addTrajectoryDirectory(os.path.join(traj_dir,
                                                             sdir))
            else:
                # Add the trajectories to the data tree.
                new_item_index = self.traj_item_tree.addTrajectoryDirectory(traj_dir)

            # Scroll the tree view so that the new item is visible and resize the
            # first column to be wide enough for the text.
            self.tvVisibleElements.scrollTo(new_item_index)
            self.tvVisibleElements.resizeColumnToContents(0)

    def selectedMapElements(self):
        """Return a list with QModelIndex objects referencing the elements
           selected in tvVisisbleElements.

        Background: For tvVisibleElements.selectedIndexes(), one element is
        a value at a row/column position. That means that, if one row is
        selected, as many indexes are returned as there are columns. However,
        in this application all values in a row belong to the same element.
        The indexes returned by selectedIndexes() for a single row all
        reference the same element, and we need only one reference per row.
        Unfortunately, the ordering of the indexes in the list returned by
        selectedIndexes() is a bit of a mess -- the elements are ordered
        'rows first' for rows selected at once by the user, and 'columns first'
        for rows selected subsequently. For example, if three rows are selected
        at once as a block (click and move down), the list will be ordered
        [r1,c1; r2,c1; .. rN,c1; r1,c2, r2,c2, ...], if the three rows are
        selected one after the other while pressing the ctrl-key, the list
        will be ordered [r1,c1; r1,c2; .. r1,cM; r2,c1, r2,c2, ...].

        The algorithm implemented in this method returns a list of the items
        represented by the rows, with only looking once at each row.
        """
        # Algorithm description: The easiest solution would be to look at each
        # element in the list returned by selectedIndexes(), however, we
        # don't need to look at each element in this list:
        #
        # Get the selected elements from the tree view (one element at each
        # row/column).
        indexes = self.tvVisibleElements.selectedIndexes()
        # Include the item referenced by the first index in the return
        # list, just make sure not to include this item again.
        returnIndexes = indexes[:1]
        # i is the current index in the list, o is an offset.
        i = 0
        o = 1
        while i < len(indexes):
            # If the item referenced by the index at i+o is another than
            # the item referenced by the index at i, include the new item
            # in the return list and look at the next index (ie. increase
            # the offset).
            if indexes[i + o].internalPointer() != indexes[i].internalPointer():
                returnIndexes.append(indexes[i + o])
                o += 1
            # If it is the same, we can use our knowledge about how the
            # elements in indexes[] are ordered: If o rows have been selected
            # in a block, we have encountered o subsequently different items.
            # We will also encounter the same elements columnCount times.
            # Thus, we can skip the next o*columnCount elements in indexes[].
            # The item referenced by the element following then will be
            # a new element, hence we can include it in returnIndexes[] and
            # continue the loop with comparing the new elements to the
            # ones following it (reset o to 1).
            else:
                i += o * self.traj_item_tree.columnCount(None)
                if i < len(indexes):
                    returnIndexes.append(indexes[i])
                o = 1
        return returnIndexes

    def selectMapElements(self):
        """Interpret the selection query entered by the user in the
           leSelectionQuery field and the select all items in the tree
           view that match the query.
        """
        # Check if the selection query should be applied to all available
        # map elements (index of cbSelectElements is 0), or only to the
        # children of the currently selected element (index is 1). In the
        # latter case also check that only one item is currently selected.
        if self.cbSelectElements.currentIndex() == 0:
            # LagrantoMapItemsTreeModel.selectionFromQuery() interprets
            # a root index of None as a selection from all available elements.
            rootIndex = None
        else:
            selectedIndexes = self.selectedMapElements()
            if len(selectedIndexes) != 1:
                QtWidgets.QMessageBox.warning(self, self.tr("select map elements"),
                                              self.tr("Please select a single element for this operation."))
                return
            else:
                rootIndex = selectedIndexes[0]

        # LagrantoMapItemsTreeModel.selectionFromQuery() returns a
        # QtGui.QItemSelection() instance containing the indexes of all
        # elements that match the query.
        #
        # FROM THE PYQT DOCUMENTATION:
        # A QItemSelection is basically a list of selection ranges, see
        # QItemSelectionRange. QItemSelection saves memory, and avoids
        # unnecessary work, by working with selection ranges rather than
        # recording the model item index for each item in the selection.
        # Generally, an instance of this class will contain a list of
        # non-overlapping selection ranges.
        itemSelection = self.traj_item_tree.selectionFromQuery(
            self.leSelectionQuery.text(),
            index=rootIndex)

        # Items can be selected in the tree view by using the select()
        # method of the tree view's selection model. select() takes either
        # a single index or an item selection:
        #     QItemSelectionModel.select (self, QModelIndex index,
        #                                 SelectionFlags command)
        #     QItemSelectionModel.select (self, QItemSelection selection,
        #                                 SelectionFlags command)
        # (see http://www.riverbankcomputing.co.uk/static/Docs/PyQt4/html/
        #   qitemselectionmodel.html#SelectionFlag-enum)
        selectionModel = self.tvVisibleElements.selectionModel()
        if itemSelection.isEmpty():
            # If no item matches the query we can't select anything.
            QtWidgets.QMessageBox.warning(self, self.tr("select map elements"),
                                          self.tr("No matching items have been found."))
        else:
            selectionModel.select(itemSelection,
                                  QtWidgets.QItemSelectionModel.ClearAndSelect |
                                  QtWidgets.QItemSelectionModel.Rows)

    def setCurrentItemColour(self):
        """Informs traj_item_tree to change the colour of the selected elements
           to the one specified by cbColour.
        """
        colour = self.cbColour.currentText()
        if colour == 'None':
            colour = "blue"
        indices = self.selectedMapElements()
        if len(indices) == 1:
            logging.debug("Changing colour of element %s",
                          indices[0].internalPointer().getName())
        else:
            logging.debug("Changing colour of selected elements")
        try:
            self.traj_item_tree.changeItemGxProperty_list(indices, 'general',
                                                          'colour', colour)
        except Exception as ex:
            QtWidgets.QMessageBox.warning(self, self.tr('item colour'), self.tr(str(ex)))

    def setCurrentItemLineStyle(self):
        """Informs traj_item_tree to change the line style of the selected
           elements to the one specified by cbLineStyle.
        """
        lineStyle = self.cbLineStyle.currentText()
        if lineStyle == 'None':
            lineStyle = "-"
        indices = self.selectedMapElements()
        if len(indices) == 1:
            logging.debug("Changing line style of element %s",
                          indices[0].internalPointer().getName())
        else:
            logging.debug("Changing line style of selected elements")
        try:
            self.traj_item_tree.changeItemGxProperty_list(indices, "general",
                                                          "linestyle", lineStyle)
        except Exception as ex:
            QtWidgets.QMessageBox.warning(self, self.tr('item line style'), self.tr("{} {}".format(type(ex), ex)))

    def setCurrentItemLineWidth(self):
        """Informs traj_item_tree to change the line width of the selected
           elements to the one specified by sbLineThickness.
        """
        lineWidth = self.sbLineWidth.value()
        indices = self.selectedMapElements()
        if len(indices) == 1:
            logging.debug("Changing line thickness of element %s",
                          indices[0].internalPointer().getName())
        else:
            logging.debug("Changing line thickness of selected elements")
        try:
            self.traj_item_tree.changeItemGxProperty_list(indices, "general",
                                                          "linewidth", lineWidth)
        except Exception as ex:
            QtWidgets.QMessageBox.warning(self, self.tr('item line width'), self.tr("{} {}".format(type(ex), ex)))

    def setCurrentItemTimeMarker(self):
        """Set the given time markers for the selected item and all its
           children (the latter for ensembles).

        The method traverses the traj_item_tree starting at the selected item.
        If markers are set for entire ensemble, user feedback is required.
        """
        first_index = None
        last_index = None
        interval = self.teTimeMarker.dateTime().toPyDateTime()
        indexStack = self.selectedMapElements()
        while len(indexStack) > 0:
            index = indexStack.pop()
            item = index.internalPointer()
            #
            # If the current item is a LagrantoOutputItem (i.e. parent for
            # a number of trajectory items) ask the user if he wants to
            # set/change the time marker for all children.
            if isinstance(item, titree.LagrantoOutputItem):
                ret = QtWidgets.QMessageBox.warning(self, self.tr("Time Marker"),
                                                    self.tr("Do you want to set the interval " +
                                                            interval.strftime("%H:%M") +
                                                            "\nfor all children of " + item.getName()),
                                                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                                    QtWidgets.QMessageBox.No)
                if ret == QtWidgets.QMessageBox.No:
                    #
                    # If the answer is no, continue with the next item on
                    # the stack.
                    continue
            #
            # If the current item has LagrantoMapItem children, push them onto
            # the stack.
            indexStack.extend([self.traj_item_tree.createIndex(child.row(), 0, child)
                               for child in item.childItems
                               if isinstance(child, titree.LagrantoMapItem)])
            #
            # Now set the time marker for the current item.
            try:
                logging.debug("Setting time marker for element %s",
                              item.getName())
                setInterval = self.traj_item_tree.setTimeMarker(index, interval,
                                                                emit_change=False)
                if not setInterval:
                    QtWidgets.QMessageBox.warning(self, self.tr("Time marker"),
                                                  self.tr("Time markers have been deleted."))
                elif setInterval != interval:
                    QtWidgets.QMessageBox.warning(self, self.tr("Time marker"),
                                                  self.tr("Warning: The minimum time interval for the"
                                                          "selected variable is " +
                                                          setInterval.strftime("%H:%M") +
                                                          ".\nThe interval has been set to this value."))
                if not first_index:
                    first_index = index
                last_index = index
            except Exception as ex:
                QtWidgets.QMessageBox.warning(self, self.tr("Time marker"), self.tr("{} {}".format(type(ex), ex)))

        self.traj_item_tree.emitChange(first_index, last_index, mode="MARKER_CHANGE")

    def plotCurrentItemInView(self):
        """
        """
        view_name = self.cbPlotInView.currentText()
        selection = self.selectedMapElements()
        if view_name != "None" and len(selection) > 0:
            view_item = (self.listviews.findItems(view_name, QtCore.Qt.MatchContains) +
                         self.listtools.findItems(view_name, QtCore.Qt.MatchContains))[0]
            logging.debug("Plotting selected elements in view <%s>", view_name)
            # Connect the trajectory tree to the view.
            view_window = view_item.window
            view = view_window.getView()
            self.traj_item_tree.setItemVisibleInView_list(
                selection, view_item.window, True)
            if view_window not in self.connected_views:
                logging.debug("Connecting to view window <%s>",
                              view_window.identifier)
                self.connected_views.append(view_window)
                view.set_trajectory_model(self.traj_item_tree)

    def removeCurrentItemFromView(self):
        """
        """
        view_name = self.cbRemoveFromView.currentText()
        if view_name != "None":
            view_item = (self.listviews.findItems(view_name, QtCore.Qt.MatchContains) +
                         self.listtools.findItems(view_name, QtCore.Qt.MatchContains))[0]
            logging.debug("Removing selected elements from view <%s>", view_name)
            self.traj_item_tree.setItemVisibleInView_list(
                self.selectedMapElements(), view_item.window, False)

    def updateViews(self):
        """Update the list of views in the comboboxes cbPlotInView and
           cbRemoveFromView.

           Connected as slot to 'viewsChanged()' of the QListWidget that
           accomodates the views.
        """
        # Remember the currently selected views in the comboboxes.
        item_plot = self.cbPlotInView.currentText()
        item_remove = self.cbRemoveFromView.currentText()

        old_views = [self.cbPlotInView.itemText(i) for i in range(self.cbPlotInView.count())]
        # Clear the boxes, add the "None" view.
        self.cbPlotInView.clear()
        self.cbRemoveFromView.clear()
        self.cbPlotInView.addItem("None")
        self.cbRemoveFromView.addItem("None")
        # Add all available views.
        for i in (self.listviews.findItems("(", QtCore.Qt.MatchContains) +
                  self.listtools.findItems("(", QtCore.Qt.MatchContains)):
            i_view = i.window.getView()
            if hasattr(i_view, "setTrajectoryModel"):
                self.cbPlotInView.addItem(i.text())
                self.cbRemoveFromView.addItem(i.text())
        # Restore the old selection (set "None" if the selected view was
        # closed).
        index = self.cbPlotInView.findText(item_plot)
        self.cbPlotInView.setCurrentIndex(index if index >= 0 else 0)
        index = self.cbRemoveFromView.findText(item_remove)
        self.cbRemoveFromView.setCurrentIndex(index if index >= 0 else 0)

        new_views = [self.cbPlotInView.itemText(i) for i in range(self.cbPlotInView.count())]
        missing_views = [str(_x) for _x in old_views if _x not in new_views]
        stack = [_x for _x in self.traj_item_tree.getRootItem().childItems]
        while len(stack) > 0:
            # Downwards traversal of the tree to determine all visible items
            # below the items that are on the stack.
            item = stack.pop()
            if hasattr(item, "views"):
                item.views = [_x for _x in item.views if _x not in missing_views]
            if hasattr(item, "childItems"):
                stack.extend(item.childItems)
        self.connected_views = [_x for _x in self.connected_views if _x not in missing_views]

    def getItemTree(self):
        """
        """
        return self.traj_item_tree



# -*- coding: utf-8 -*-
"""

    mslib.mscolab.chat_manager
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Code to handle socket connections in mscolab

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from mslib.mscolab.models import db, Message, User


class ChatManager(object):
    """Class with handler functions for chat related functionalities"""

    def __init__(self):
        pass

    def add_message(self, user, text, roomname):
        """
        text: message to be emitted to room and saved to db
        roomname: room-name(p_id) to which message is emitted,
        user: User object, one which emits the message
        """
        message = Message(roomname, user.id, text)
        db.session.add(message)
        db.session.commit()
        return message

    def get_messages(self, p_id, last_timestamp=None):
        """
        p_id: project id
        last_timestamp:  if provided, messages only after this time stamp is provided
        """
        messages = Message.query.filter_by(p_id=p_id)
        if last_timestamp:
            messages = messages.filter(Message.created_at > last_timestamp)
        messages = messages.all()

        messages = list(map(lambda x:
                        {'user': x.u_id,
                         'time': x.created_at.strftime("%m %d %Y, %H:%M:%S"),
                         'text': x.text,
                         'username': self.get_user_from_id(x.u_id).username}, messages))
        return messages

    # refactor to utils
    def get_user_from_id(self, id):
        return User.query.filter_by(id=id).first()



# -*- coding: utf-8 -*-
"""

    mslib.mscolab.add_bulk_permissions
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This file is used to insert permissions in bulk to mscolab db. This would in future, be
    evolved to a circle management module, to deal with groups of users(circles).

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from flask import Flask
import logging
import sys

from mslib.mscolab.models import User, db, Project, Permission
from mslib.mscolab.conf import mscolab_settings

logging.getLogger().setLevel(logging.INFO)
# set the project root directory as the static folder
app = Flask(__name__, static_url_path='')
app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
app.config['SECRET_KEY'] = mscolab_settings.SECRET_KEY

app = Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
app.config['SECRET_KEY'] = 'secret!'.encode('utf-8')
db.init_app(app)
path = sys.argv[1]


def main():
    with app.app_context():
        file = open(path, 'r')
        content = file.read()
        groups = content.split('\n\n')
        for group in groups:
            group = group.split('\n')
            p = Project.query.filter_by(path=group[0]).first()
            if p is None:
                logging.info("Project {} doesn't exist, please create it.")
                continue
            users = group[1:]
            for user in users:
                _username = user.split('-')[0]
                permission = user.split('-')[1]
                if permission == 'a':
                    permission = 'admin'
                elif permission == 'c':
                    permission = 'collaborator'
                else:
                    permission = 'viewer'

                u = User.query.filter_by(username=_username).first()
                if u is None:
                    logging.info("User {} doesn't exist, skipping entry".format(_username))
                    continue
                perm_existing = Permission.query.filter_by(u_id=u.id, p_id=p.id).first()
                if perm_existing:
                    logging.info("permission exists, skipping entry")
                    continue
                perm = Permission(u.id, p.id, permission)
                db.session.add(perm)
                db.session.commit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

    mslib.mscolab.demodata
    ~~~~~~~~~~~~~~~~~~~~~~

    dummydata for mscolab

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import fs
import sys
from flask import Flask
import logging
import argparse
import git
import psycopg2
import sqlalchemy
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT


try:
    import MySQLdb as ms
except ImportError:
    ms = None
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import User, Project, Permission
from mslib.msui import MissionSupportSystemDefaultConfig as mss_default
from mslib.mscolab.seed import seed_data, create_tables
from fs.tempfs import TempFS

ROOT_FS = TempFS(identifier="mss")
ROOT_DIR = ROOT_FS.root_path


def create_test_data():
    # for tempfile_mscolab.ftml
    create_mssdir()
    # creating test directory
    fs_datadir = fs.open_fs(mscolab_settings.BASE_DIR)
    if fs_datadir.exists('colabdata'):
        fs_datadir.removetree('colabdata')
    fs_datadir.makedir('colabdata')
    fs_datadir = fs.open_fs(mscolab_settings.DATA_DIR)
    # creating filedata directory
    create_test_files()

    if mscolab_settings.SQLALCHEMY_DB_URI.split(':')[0] == "mysql":
        if ms is None:
            logging.info("""can't complete demodata setup,
                         use sqlite3 or configure mysql with proper modules""")
            sys.exit(0)
        try:
            db = ms.connect(host=mscolab_settings.DB_HOST,    # your host, usually localhost
                            user=mscolab_settings.DB_USER,         # your username
                            passwd=mscolab_settings.DB_PASSWORD,  # your password
                            db=mscolab_settings.DB_NAME)        # name of the data base
            cursor = db.cursor()
            logging.info("Database exists, please drop it before running mscolab/demodata.py")
            sys.exit(0)
        except Exception as e:
            logging.debug(e)
            db = ms.connect(host=mscolab_settings.DB_HOST, user=mscolab_settings.DB_USER,
                            passwd=mscolab_settings.DB_PASSWORD)
            cursor = db.cursor()
            sql = 'CREATE DATABASE ' + mscolab_settings.DB_NAME + ';'
            cursor.execute(sql)
            db = ms.connect(host=mscolab_settings.DB_HOST,    # your host, usually localhost
                            user=mscolab_settings.DB_USER,         # your username
                            passwd=mscolab_settings.DB_PASSWORD,  # your password
                            db=mscolab_settings.DB_NAME)        # name of the data base
            cursor = db.cursor()

        PATH_TO_FILE = os.getcwd() + '/schema_seed.sql'
        for line in open(PATH_TO_FILE):
            if line.split(' ')[0] not in ['CREATE', 'SET']:
                continue
            cursor.execute(line)

        app = Flask(__name__)
        app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
        from mslib.mscolab.models import db
        app.config['SECRET_KEY'] = 'secret!'.encode('utf-8')
        db.init_app(app)
        with app.app_context():
            db.create_all()
            data = [
                ('a', 8, 'a', 'a'),
                ('b', 9, 'b', 'b'),
                ('c', 10, 'c', 'c'),
            ]
            for data_point in data:
                user = User(data_point[0], data_point[3], data_point[2])
                user.id = data_point[1]
                db.session.add(user)
                db.session.commit()

            data = [
                (1, 'one', 'a, b'),
                (2, 'two', 'b, c'),
                (3, 'three', 'a, c'),
            ]
            for data_point in data:
                project = Project(data_point[1], data_point[2])
                project.id = data_point[0]
                db.session.add(project)
            db.session.commit()

            data = [
                (1, 8, 1, 'creator'),
                (2, 9, 1, 'collaborator'),
                (3, 9, 2, 'creator'),
                (4, 10, 2, 'collaborator'),
                (5, 10, 3, 'creator'),
                (6, 8, 3, 'collaborator'),
                (7, 10, 1, 'viewer'),
            ]
            for data_point in data:
                project = Permission(data_point[1], data_point[2], data_point[3])
                db.session.add(project)
            db.session.commit()

        pass
    elif mscolab_settings.SQLALCHEMY_DB_URI.split(':')[0] == "sqlite":
        create_tables(mscolab_settings.SQLALCHEMY_DB_URI)
        seed_data(mscolab_settings.SQLALCHEMY_DB_URI)

    elif mscolab_settings.SQLALCHEMY_DB_URI.split(':')[0] == "postgresql":
        create_postgres_test()


def create_test_config():
    config_string = '''# -*- coding: utf-8 -*-
"""

    mslib.mscolab.conf.py.example
    ~~~~~~~~~~~~~~~~~~~~

    config for mscolab.

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
# SQLALCHEMY_DB_URI = 'mysql://user:pass@127.0.0.1/mscolab'
import os
import logging
import fs

from fs.tempfs import TempFS


ROOT_FS = TempFS(identifier="mss")
ROOT_DIR = ROOT_FS.root_path


# directory where mss output files are stored
root_fs = fs.open_fs(ROOT_DIR)
root_fs.makedir('colabdata')
DATA_DIR = os.path.join(ROOT_DIR, 'colabdata')
BASE_DIR = ROOT_DIR
SQLITE_FILE_PATH = os.path.join(DATA_DIR, 'mscolab.db')

SQLALCHEMY_DB_URI = 'sqlite:///' + SQLITE_FILE_PATH

# used to generate and parse tokens
# details of database connections
SECRET_KEY = 'secretkEyu'
DB_HOST = '127.0.0.1'
DB_USER = 'user'
DB_PASSWORD = 'pass'
DB_NAME = 'test_1'

# SQLALCHEMY_DB_URI = 'postgresql://{}:{}@{}/{}'.format(DB_USER, DB_PASSWORD, DB_HOST, DB_NAME)

# mscolab data directory
MSCOLAB_DATA_DIR = os.path.join(DATA_DIR, 'filedata')
# text to be written in new mscolab based ftml files.
STUB_CODE = """<?xml version="1.0" encoding="utf-8"?>
<FlightTrack version="1.7.6">
  <ListOfWaypoints>
    <Waypoint flightlevel="250" lat="67.821" location="Kiruna" lon="20.336">
      <Comments></Comments>
    </Waypoint>
    <Waypoint flightlevel="250" lat="78.928" location="Ny-Alesund" lon="11.986">
      <Comments></Comments>
    </Waypoint>
  </ListOfWaypoints>
</FlightTrack>
"""
    '''
    ROOT_FS = fs.open_fs(ROOT_DIR)
    if not ROOT_FS.exists('mscolab'):
        ROOT_FS.makedir('mscolab')
    mscolab_fs = fs.open_fs(os.path.join(ROOT_DIR, "mscolab"))
    mscolab_fs.writetext('mscolab_settings.py', config_string)
    mscolab_fs.close()
    return (os.path.join(ROOT_DIR, 'mscolab', 'mscolab_settings.py'),
            os.path.join(ROOT_DIR, 'mscolab'))


def create_test_files():
    fs_datadir = fs.open_fs(mscolab_settings.DATA_DIR)
    if not fs_datadir.exists('filedata'):
        fs_datadir.makedir('filedata')
        # add files
        file_dir = fs.open_fs(fs.path.combine(mscolab_settings.BASE_DIR, 'colabdata/filedata'))
        # make directories
        file_paths = ['one', 'two', 'three']
        for file_path in file_paths:
            file_dir.makedir(file_path)
            file_dir.writetext('{}/main.ftml'.format(file_path), mscolab_settings.STUB_CODE)
            # initiate git
            r = git.Repo.init(fs.path.combine(mscolab_settings.BASE_DIR,
                                              'colabdata/filedata/{}'.format(file_path)))
            r.index.add(['main.ftml'])
            r.index.commit("initial commit")
        file_dir.close()


def create_postgres(seed=False):
    try:
        # if database exists it'll create tables
        create_tables(mscolab_settings.SQLALCHEMY_DB_URI)
    except sqlalchemy.exc.OperationalError as e:
        if e.args[0].find("database \"{}\" does not exist".format(mscolab_settings.DB_NAME)) != -1:
            logging.debug("database doesn't exist, creating one")
            con = psycopg2.connect(dbname="template1",
                                   user=mscolab_settings.DB_USER,
                                   host=mscolab_settings.DB_HOST,
                                   password=mscolab_settings.DB_PASSWORD)

            con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

            cur = con.cursor()
            cur.execute("CREATE DATABASE {};".format(mscolab_settings.DB_NAME))
            create_tables(mscolab_settings.SQLALCHEMY_DB_URI)
            if seed:
                seed_data(mscolab_settings.SQLALCHEMY_DB_URI)


def create_postgres_test():
    con = psycopg2.connect(dbname="template1",
                           user=mscolab_settings.DB_USER,
                           host=mscolab_settings.DB_HOST,
                           password=mscolab_settings.DB_PASSWORD)

    con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

    cur = con.cursor()
    cur.execute("DROP DATABASE IF EXISTS {};".format(mscolab_settings.DB_NAME))
    cur.execute("CREATE DATABASE {};".format(mscolab_settings.DB_NAME))
    create_tables(mscolab_settings.SQLALCHEMY_DB_URI)
    # to reset cursors
    con = psycopg2.connect(dbname=mscolab_settings.DB_NAME,
                           user=mscolab_settings.DB_USER,
                           host=mscolab_settings.DB_HOST,
                           password=mscolab_settings.DB_PASSWORD)

    con.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

    cur = con.cursor()
    cur.execute("ALTER SEQUENCE users_id_seq RESTART WITH 4;")
    cur.execute("ALTER SEQUENCE projects_id_seq RESTART WITH 4;")
    cur.execute("ALTER SEQUENCE permissions_id_seq RESTART WITH 11;")
    seed_data(mscolab_settings.SQLALCHEMY_DB_URI)


def create_data():
    create_mssdir()
    fs_datadir = fs.open_fs(mscolab_settings.BASE_DIR)
    if not fs_datadir.exists('colabdata'):
        fs_datadir.makedir('colabdata')
    fs_datadir = fs.open_fs(mscolab_settings.DATA_DIR)
    if not fs_datadir.exists('filedata'):
        fs_datadir.makedir('filedata')
    if mscolab_settings.SQLALCHEMY_DB_URI.split(':')[0] == "sqlite":
        # path_prepend = os.path.dirname(os.path.abspath(__file__))
        create_tables(mscolab_settings.SQLALCHEMY_DB_URI)
    elif mscolab_settings.SQLALCHEMY_DB_URI.split(':')[0] == "postgresql":
        create_postgres()


def create_mssdir():
    fs_datadir = fs.open_fs('~')
    basename = fs.path.basename(mss_default.mss_dir)
    if not fs_datadir.exists(basename):
        fs_datadir.makedir(basename)


def main():
    parser = argparse.ArgumentParser(description="Tool to setup data for usage of mscolab")
    parser.add_argument("--test", action="store_true", help="setup test data")
    parser.add_argument("--init", action="store_true", help="setup deployment data")
    args = parser.parse_args()
    if args.test:
        create_test_data()
    elif args.init:
        create_data()
    else:
        print("for help, use -h flag")


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""

    mslib.msui.socket_control
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    client socket connection handler

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import socketio
import json
import logging

from mslib.msui.mss_qt import QtCore
from mslib.msui import MissionSupportSystemDefaultConfig as mss_default


class ConnectionManager(QtCore.QObject):

    signal_reload = QtCore.Signal(int, name="reload_wps")
    signal_autosave = QtCore.Signal(int, int, name="autosave en/db")
    signal_message_receive = QtCore.Signal(str, str, name="message rcv")
    signal_new_permission = QtCore.Signal(int, int, name="new permission")
    signal_update_permission = QtCore.Signal(int, int, str, name="update permission")
    signal_revoke_permission = QtCore.Signal(int, int, name="revoke permission")

    def __init__(self, token, user, mscolab_server_url=mss_default.mscolab_server_url):
        super(ConnectionManager, self).__init__()
        self.sio = socketio.Client(reconnection_attempts=5)
        self.sio.on('file-changed', handler=self.handle_file_change)
        # ToDo merge them into one 'autosave-client' event
        self.sio.on('autosave-client-en', handler=self.handle_autosave_enable)
        self.sio.on('autosave-client-db', handler=self.handle_autosave_disable)
        # on chat message recive
        self.sio.on('chat-message-client', handler=self.handle_incoming_message)
        # on new permission
        self.sio.on('new-permission', handler=self.handle_new_permission)
        # on update of permission
        self.sio.on('update-permission', handler=self.handle_update_permission)
        # on revoking project permission
        self.sio.on('revoke-permission', handler=self.handle_revoke_permission)
        self.mscolab_server_url = mscolab_server_url
        self.sio.connect(self.mscolab_server_url)
        self.sio.emit('start', {'token': token})
        self.token = token
        self.user = user

    def handle_update_permission(self, message):
        """
        signal update of permission affected
        """
        message = json.loads(message)
        p_id = int(message["p_id"])
        u_id = int(message["u_id"])
        access_level = message["access_level"]
        self.signal_update_permission.emit(p_id, u_id, access_level)

    def handle_new_permission(self, message):
        """
        signal updating of newly added permission
        """
        message = json.loads(message)
        p_id = int(message["p_id"])
        u_id = int(message["u_id"])
        self.signal_new_permission.emit(p_id, u_id)

    def handle_revoke_permission(self, message):
        """
        Signal update of revoked permission
        """
        message = json.loads(message)
        p_id = int(message["p_id"])
        u_id = int(message["u_id"])
        self.signal_revoke_permission.emit(p_id, u_id)

    def handle_incoming_message(self, message):
        # raise signal to render to view
        message = json.loads(message)
        logging.debug(message)
        # emit signal
        self.signal_message_receive.emit(message["user"], message["message_text"])

    def handle_file_change(self, message):
        message = json.loads(message)
        self.signal_reload.emit(message["p_id"])

    def handle_new_room(self, p_id):
        logging.debug("adding user to new room")
        self.sio.emit('add-user-to-room', {
                      "p_id": p_id,
                      "token": self.token})

    def send_message(self, message_text, p_id):
        logging.debug("sending message")
        self.sio.emit('chat-message', {
                      "p_id": p_id,
                      "token": self.token,
                      "message_text": message_text})

    def save_file(self, token, p_id, content, comment=None):
        logging.debug("saving file")
        self.sio.emit('file-save', {
                      "p_id": p_id,
                      "token": self.token,
                      "content": content,
                      "comment": comment})

    def emit_autosave(self, token, p_id, enable):
        logging.debug("emitting autosave")
        self.sio.emit('autosave', {
                      "p_id": p_id,
                      "token": token,
                      "enable": enable})

    # ToDo directly call self.signal_autosave

    def handle_autosave_enable(self, message):
        message = json.loads(message)
        self.signal_autosave.emit(1, message["p_id"])

    def handle_autosave_disable(self, message):
        message = json.loads(message)
        self.signal_autosave.emit(0, message["p_id"])

    def disconnect(self):
        self.sio.disconnect()



# -*- coding: utf-8 -*-
"""

    mslib.msui.mscolab_project
    ~~~~~~~~~~~~~~~~~~~~~

    Mscolab project window, to display chat, file change

    This file is part of mss.

    :copyright: 2019 Shivashis Padhi
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from mslib.msui.mss_qt import QtCore, QtWidgets
from mslib.msui.mss_qt import ui_mscolab_project_window as ui
from mslib.msui import MissionSupportSystemDefaultConfig as mss_default
from mslib.utils import config_loader

import logging
import requests
import json
import datetime


class MSColabProjectWindow(QtWidgets.QMainWindow, ui.Ui_MscolabProject):
    """Derives QMainWindow to provide some common functionality to all
       MSUI view windows.
    """
    name = "MSColab Project Window"
    identifier = None

    viewCloses = QtCore.pyqtSignal(name="viewCloses")
    reloadWindows = QtCore.pyqtSignal(name="reloadWindows")

    def __init__(self, token, p_id, conn, access_level, parent=None,
                 mscolab_server_url=config_loader(dataset="default_MSCOLAB", default=mss_default.default_MSCOLAB)):
        """
        token: access_token
        p_id: project id
        conn: to send messages, recv messages, if a direct slot-signal can't be setup
            to be connected at parents'
        """
        super(MSColabProjectWindow, self).__init__(parent)
        self.setupUi(self)
        # set url
        self.mscolab_server_url = mscolab_server_url

        # constrain vertical layout
        # self.verticalLayout.setSizeConstraint(self.verticalLayout_6.SetMinimumSize)
        # self.messages.setWordWrap(True)

        self.token = token
        self.p_id = p_id
        self.conn = conn
        self.access_level = access_level
        # add receive message handler
        self.conn.signal_message_receive.connect(self.render_new_message)
        logging.debug(ui.Ui_MscolabProject)
        # establish button press handlers
        self.add.clicked.connect(self.add_handler)
        self.modify.clicked.connect(self.modify_handler)
        self.delete_1.clicked.connect(self.delete_handler)
        self.checkout.clicked.connect(self.handle_undo)
        # send message handler
        self.sendMessage.clicked.connect(self.send_message)
        # load users
        self.load_users()
        # load changes
        self.load_all_changes()
        # load messages
        self.load_all_messages()
        # active change
        self.active_ch_id = None
        user_d = self.get_user_details(self.token)
        proj_d = self.get_project_details(self.p_id)
        self.user_info.setText("logged in as: {}".format(user_d["username"]))
        self.proj_info.setText("Project name: {}".format(proj_d["path"]))
        # disable admin actions if viewer or collaborator
        self.check_permission_and_render_control(self.access_level)

    def check_permission_and_render_control(self, access_level):
        """
        block some control according to access_level
        """
        if access_level == "collaborator" or access_level == "viewer":
            self.add.setEnabled(False)
            self.modify.setEnabled(False)
            self.delete_1.setEnabled(False)
        else:
            self.add.setEnabled(True)
            self.modify.setEnabled(True)
            self.delete_1.setEnabled(True)

        if access_level == "viewer":
            self.checkout.setEnabled(False)
            self.sendMessage.setEnabled(False)
        else:
            self.checkout.setEnabled(True)
            self.sendMessage.setEnabled(True)

    def get_user_details(self, token):
        """
        token: auth token of the user to be fetched from mscolab server
        """
        data = {
            'token': token
        }
        r = requests.get(self.mscolab_server_url + '/user', data=data)
        json_ = json.loads(r.text)
        return json_['user']

    def get_project_details(self, p_id):
        """
        p_id: active project id details of which are fetched from mscolab server
        """
        data = {
            'token': self.token,
            'p_id': p_id
        }
        r = requests.get(self.mscolab_server_url + '/project_details', data=data)
        json_ = json.loads(r.text)
        return json_

    def send_message(self):
        """
        send message through connection
        """
        message_text = self.messageText.toPlainText()
        self.conn.send_message(message_text, self.p_id)
        self.messageText.clear()

    def add_handler(self):
        # get username, p_id
        username = self.username.text()
        access_level = str(self.accessLevel.currentText())
        # send request to add
        data = {
            "token": self.token,
            "p_id": self.p_id,
            "username": username,
            "access_level": access_level
        }
        r = requests.post(self.mscolab_server_url + '/add_permission', data=data)
        if r.text == "True":
            self.load_users()
        else:
            # show a QMessageBox with errors
            pass

    def modify_handler(self):
        # get username, p_id
        username = self.username.text()
        access_level = str(self.accessLevel.currentText())
        # send request to modify
        data = {
            "token": self.token,
            "p_id": self.p_id,
            "username": username,
            "access_level": access_level
        }
        r = requests.post(self.mscolab_server_url + '/modify_permission', data=data)
        if r.text == "True":
            self.load_users()
        else:
            # show a QMessageBox with errors
            pass

    def delete_handler(self):
        # get username, p_id
        username = self.username.text()
        p_id = self.p_id
        # send request to delete
        data = {
            "token": self.token,
            "p_id": p_id,
            "username": username
        }
        r = requests.post(self.mscolab_server_url + '/revoke_permission', data=data)
        if r.text == "True":
            self.load_users()
        else:
            # show a QMessageBox with errors
            pass

    def load_users(self):
        # load users to side-tab here

        # make request to get users
        data = {
            "token": self.token,
            "p_id": self.p_id
        }
        r = requests.get(self.mscolab_server_url + '/authorized_users', data=data)
        if r.text == "False":
            # show QMessageBox errors here
            pass
        else:
            self.collaboratorsList.clear()
            users = json.loads(r.text)["users"]
            for user in users:
                item = QtWidgets.QListWidgetItem("{} - {}".format(user["username"],
                                                 user["access_level"]),
                                                 parent=self.collaboratorsList)
                item.username = user["username"]
                self.collaboratorsList.addItem(item)
            self.collaboratorsList.itemActivated.connect(self.update_username_wrt_item)

    def update_username_wrt_item(self, item):
        self.username.setText(item.username)

    @QtCore.Slot(str, str)
    def render_new_message(self, username, message, scroll=True):
        item = QtWidgets.QListWidgetItem("{}: {}\n".format(username, message), parent=self.messages)
        self.messages.addItem(item)
        if scroll:
            self.messages.scrollToBottom()

    def load_all_changes(self):
        """
        get changes from api, clear listwidget, render them to ui
        """
        data = {
            "token": self.token,
            "p_id": self.p_id
        }
        # 'get all changes' request
        r = requests.get(self.mscolab_server_url + '/get_changes', data=data)
        changes = json.loads(r.text)["changes"]
        self.changes.clear()
        self.active_ch_id = None
        for change in changes:
            item = QtWidgets.QListWidgetItem("{}: {}\n".format(change["username"],
                                             change["content"]), parent=self.changes)
            item._id = change['id']
            self.changes.addItem(item)
        self.changes.scrollToBottom()

    def handle_undo(self):
        index = self.changes.currentIndex()
        qm = QtWidgets.QMessageBox
        if not index.isValid():
            qm.critical(
                self, self.tr("Undo"),
                self.tr("Please select a change first."))
        else:
            ret = qm.question(
                self, self.tr("Undo"),
                "Do you want to checkout to this change?", qm.Yes, qm.No)

            if ret == qm.Yes:
                self.request_undo_mscolab(index)

    def request_undo_mscolab(self, index):
        # undo change from server
        data = {
            "token": self.token,
            "ch_id": self.changes.itemFromIndex(index)._id
        }
        # 'undo' request
        r = requests.post(self.mscolab_server_url + '/undo', data=data)
        if r.text == "True":
            # reload windows
            self.reloadWindows.emit()

    def load_all_messages(self):
        # empty messages and reload from server
        data = {
            "token": self.token,
            "p_id": self.p_id,
            "timestamp": datetime.datetime(1970, 1, 1).strftime("%m %d %Y, %H:%M:%S")
        }
        # returns an array of messages
        r = requests.post(self.mscolab_server_url + "/messages", data=data)
        response = json.loads(r.text)

        messages = response["messages"]
        logging.debug(messages)
        logging.debug(type(messages))
        # clear message box
        self.messages.clear()
        for message in messages:
            username = message["username"]
            message_text = message["text"]
            self.render_new_message(username, message_text, scroll=False)
        self.messages.scrollToBottom()

    def closeEvent(self, event):
        self.viewCloses.emit()



# -*- coding: utf-8 -*-
"""

    mslib.mscolab.file_manager.py
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Code to handle file I/O in mscolab

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import fs
import difflib
import logging
import git
from mslib.mscolab.models import db, Project, Permission, User, Change, Message
from mslib.mscolab.conf import mscolab_settings


class FileManager(object):
    """Class with handler functions for file related functionalities"""

    def __init__(self, data_dir):
        self.data_dir = data_dir

    def create_project(self, path, description, user, content=None):
        """
        path: path to the project
        description: description of the project
        """
        # set codes on these later
        if path.find("/") != -1 or path.find("\\") != -1 or (" " in path):
            logging.debug("malicious request: %s", user)
            return False
        proj_available = Project.query.filter_by(path=path).first()
        if proj_available:
            return False
        project = Project(path, description, False)
        db.session.add(project)
        db.session.flush()
        project_id = project.id
        # this is the only insertion with "creator" access_level
        perm = Permission(user.id, project_id, "creator")
        db.session.add(perm)
        db.session.commit()
        data = fs.open_fs(self.data_dir)
        data.makedir(project.path)
        project_file = data.open(fs.path.combine(project.path, 'main.ftml'), 'w')
        if content is not None:
            project_file.write(content)
        else:
            project_file.write(mscolab_settings.STUB_CODE)
        project_path = fs.path.combine(self.data_dir, project.path)
        r = git.Repo.init(project_path)
        r.index.add(['main.ftml'])
        r.index.commit("initial commit")
        return True

    def get_project_details(self, p_id, user):
        """
        p_id: project id
        user: authenticated user
        """
        project = Project.query.filter_by(id=p_id).first()
        project = {"id": project.id,
                   "path": project.path,
                   "description": project.description,
                   "autosave": project.autosave
                   }
        return project

    def add_permission(self, p_id, u_id, username, access_level, user):
        """
        p_id: project id
        u_id: user-id who is being given permission
        access_level: the access level given to user
        user: authorized user, making this request
        """
        if not self.is_admin(user.id, p_id):
            return False
        if username:
            user_victim = User.query.filter_by(username=username).first()
            if not user_victim:
                return False
            u_id = user_victim.id
        perm_old = Permission.query.filter_by(u_id=u_id, p_id=p_id).first()
        if perm_old or (access_level == "creator"):
            return False
        perm_new = Permission(u_id, p_id, access_level)
        db.session.add(perm_new)
        db.session.commit()
        return True

    def revoke_permission(self, p_id, u_id, username, user):
        """
        p_id: project id
        u_id: user-id
        username: to identify victim user
        user: logged in user
        """
        deleted = None
        if user.id == u_id:
            return False
        if not self.is_admin(user.id, p_id):
            return False
        else:
            if username:
                user_victim = User.query.filter_by(username=username).first()
                if not user_victim:
                    return False
                u_id = user_victim.id
            perm = Permission.query.filter_by(u_id=u_id, p_id=p_id).first()
            if perm is None:
                return False
            if perm.access_level == "creator":
                return False
            deleted = Permission.query.filter_by(u_id=u_id, p_id=p_id).delete()
            db.session.commit()
        if deleted:
            return True
        else:
            return False

    def list_projects(self, user):
        """
        user: logged in user
        """
        projects = []
        permissions = Permission.query.filter_by(u_id=user.id).all()
        for permission in permissions:
            project = Project.query.filter_by(id=permission.p_id).first()
            projects.append({
                            "p_id": permission.p_id,
                            "access_level": permission.access_level,
                            "path": project.path,
                            "description": project.description,
                            "autosave": project.autosave})
        return projects

    def is_admin(self, u_id, p_id):
        """
        p_id: project id
        u_id: user-id
        """
        # return true only if the user is admin or creator
        perm = Permission.query.filter_by(u_id=u_id, p_id=p_id).first()
        if not perm:
            return False
        elif perm.access_level != "admin" and perm.access_level != "creator":
            return False
        return True

    def auth_type(self, u_id, p_id):
        """
        p_id: project id
        u_id: user-id
        """
        perm = Permission.query.filter_by(u_id=u_id, p_id=p_id).first()
        if not perm:
            return False
        return perm.access_level

    def update_project(self, p_id, attribute, value, user):
        """
        p_id: project id
        attribute: attribute to be changed, eg path
        user: logged in user
        """
        if not self.is_admin(user.id, p_id):
            return False
        project = Project.query.filter_by(id=p_id).first()
        if attribute == "path":
            if value.find("/") != -1 or value.find("\\") != -1 or (" " in value):
                logging.debug("malicious request: %s", user)
                return False
            data = fs.open_fs(self.data_dir)
            if data.exists(value):
                return False
            # will be move when projects are introduced
            # make a directory, else movedir fails
            data.makedir(value)
            data.movedir(project.path, value)
        setattr(project, attribute, value)
        db.session.commit()
        return True

    def update_access_level(self, p_id, u_id, username, access_level, user):
        if not self.is_admin(user.id, p_id):
            return False
        if username:
            user_victim = User.query.filter_by(username=username).first()
            if not user_victim:
                return False
            u_id = user_victim.id
        if u_id is None or u_id == user.id:
            return
        perm = Permission.query.filter_by(u_id=u_id, p_id=p_id).first()
        if not perm or perm.access_level == "creator":
            return False
        perm.access_level = access_level
        db.session.commit()
        return True

    def delete_file(self, p_id, user):
        """
        p_id: project id
        user: logged in user
        """
        if self.auth_type(user.id, p_id) != "creator":
            return False
        Permission.query.filter_by(p_id=p_id).delete()
        Change.query.filter_by(p_id=p_id).delete()
        Message.query.filter_by(p_id=p_id).delete()
        Change.query.filter_by(p_id=p_id).delete()
        project = Project.query.filter_by(id=p_id).first()
        data = fs.open_fs(self.data_dir)
        data.removetree(project.path)
        project = Project.query.filter_by(id=p_id).delete()
        db.session.commit()
        return True

    def get_authorized_users(self, p_id):
        """
        p_id: project-id
        """
        permissions = Permission.query.filter_by(p_id=p_id).all()
        users = []
        for permission in permissions:
            user = User.query.filter_by(id=permission.u_id).first()
            users.append({"username": user.username, "access_level": permission.access_level})
        return users

    def save_file(self, p_id, content, user, comment=""):
        """
        p_id: project-id,
        content: content of the file to be saved
        # ToDo save change in schema
        """
        project = Project.query.filter_by(id=p_id).first()
        if not project:
            return False
        data = fs.open_fs(self.data_dir)
        """
        old file is read, the diff between old and new is calculated and stored
        as 'Change' in changes table. comment for each change is optional
        """
        project_file = data.open(fs.path.combine(project.path, 'main.ftml'), 'r')
        old_data = project_file.read()
        project_file.close()
        old_data_lines = old_data.splitlines()
        content_lines = content.splitlines()
        diff = difflib.unified_diff(old_data_lines, content_lines, lineterm='')
        diff_content = '\n'.join(list(diff))
        project_file = data.open(fs.path.combine(project.path, 'main.ftml'), 'w')
        project_file.write(content)
        project_file.close()
        # commit changes if comment is not None
        if diff_content != "":
            # commit to git repository
            project_path = fs.path.combine(self.data_dir, project.path)
            repo = git.Repo(project_path)
            repo.index.add(['main.ftml'])
            # hack used, ToDo fix it
            if comment == "" or comment is False or comment is None:
                comment = "committing change"
            cm = repo.index.commit(comment)
            # change db table
            change = Change(p_id, user.id, diff_content, cm.hexsha, comment)
            db.session.add(change)
            db.session.commit()
        return True

    def get_file(self, p_id, user):
        """
        p_id: project-id
        user: user of this request
        """
        perm = Permission.query.filter_by(u_id=user.id, p_id=p_id).first()
        if not perm:
            return False
        project = Project.query.filter_by(id=p_id).first()
        if not project:
            return False
        data = fs.open_fs(self.data_dir)
        project_file = data.open(fs.path.combine(project.path, 'main.ftml'), 'r')
        return project_file.read()

    def get_changes(self, p_id, user):
        """
        p_id: project-id
        user: user of this request

        Get all changes, mostly to be used in the chat window, in the side panel
        to render the recent changes.
        """
        perm = Permission.query.filter_by(u_id=user.id, p_id=p_id).first()
        if not perm:
            return False
        changes = Change.query.filter_by(p_id=p_id).all()
        return list(map(lambda change: {'content': change.content,
                                        'comment': change.comment,
                                        'u_id': change.u_id,
                                        'username': self.get_user_from_id(change.u_id).username,
                                        'id': change.id}, changes,))

    def get_user_from_id(self, id):
        return User.query.filter_by(id=id).first()

    def get_change_by_id(self, ch_id, user):
        """
        ch_id: change id
        user: user of this request

        Get change related to id
        """
        change = Change.query.filter_by(id=ch_id).first()
        if not change:
            return False
        perm = Permission.query.filter_by(u_id=user.id, p_id=change.p_id).first()
        if not perm:
            return False
        return {'content': change.content, 'comment': change.comment, 'u_id': change.u_id}

    def undo(self, ch_id, user):
        """
        ch_id: change-id
        user: user of this request

        Undo a change
        # ToDo a revert option, which removes only that commit's change
        """
        ch = Change.query.filter_by(id=ch_id).first()
        if ch is None:
            return False
        project = Project.query.filter_by(id=ch.p_id).first()
        if not ch or not project:
            return False

        data = fs.open_fs(self.data_dir)
        project_file = data.open(fs.path.combine(project.path, 'main.ftml'), 'r')
        old_data = project_file.read()
        project_file.close()
        old_data_lines = old_data.splitlines()

        project_path = fs.path.combine(self.data_dir, project.path)
        repo = git.Repo(project_path)
        try:
            file_content = repo.git.show('{}:{}'.format(ch.commit_hash, 'main.ftml'))

            content_lines = file_content.splitlines()
            diff = difflib.unified_diff(old_data_lines, content_lines, lineterm='')
            diff_content = '\n'.join(list(diff))

            proj_fs = fs.open_fs(project_path)
            proj_fs.writetext('main.ftml', file_content)
            proj_fs.close()
            repo.index.add(['main.ftml'])
            cm = repo.index.commit("checkout to {}".format(ch.commit_hash))

            change = Change(ch.p_id, user.id, diff_content, cm.hexsha,
                            "checkout to {}".format(ch.commit_hash))
            db.session.add(change)
            db.session.commit()

            return True
        except Exception as ex:
            logging.debug(ex)
            return False



# -*- coding: utf-8 -*-
"""

    mslib.msui._tests.test_mscolab_project
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    This module is used to test mscolab-project related gui.

    This file is part of mss.

    :copyright: Copyright 2019 Shivashis Padhi
    :copyright: Copyright 2019-2020 by the mss team, see AUTHORS.
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import sys
from mslib.msui.mss_qt import QtWidgets, QtTest, QtCore
import logging
import time

from mslib.mscolab.server import db, app, initialize_managers
from mslib._tests.constants import MSCOLAB_URL_TEST
from mslib.mscolab.conf import mscolab_settings
from mslib.mscolab.models import Message, Change
import mslib.msui.mscolab as mc


class Test_MscolabProject(object):
    def setup(self):

        # start mscolab server
        self.app = app
        self.app.config['SQLALCHEMY_DATABASE_URI'] = mscolab_settings.SQLALCHEMY_DB_URI
        self.app.config['MSCOLAB_DATA_DIR'] = mscolab_settings.MSCOLAB_DATA_DIR
        self.app, _, cm, fm = initialize_managers(self.app)
        self.fm = fm
        self.cm = cm
        db.init_app(self.app)

        logging.debug("starting")
        self.application = QtWidgets.QApplication(sys.argv)
        self.window = mc.MSSMscolabWindow(data_dir=mscolab_settings.MSCOLAB_DATA_DIR,
                                          mscolab_server_url=MSCOLAB_URL_TEST)
        self._login()
        self._activate_project_at_index(0)
        # activate project window here by clicking button
        QtTest.QTest.mouseClick(self.window.projWindow, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        self.proj_window = self.window.project_window
        QtTest.QTest.qWaitForWindowExposed(self.window)
        QtWidgets.QApplication.processEvents()

    def teardown(self):
        # to disconnect connections, and clear token
        # Not logging out since it pops up a dialog
        # self.window.logout()
        for window in self.window.active_windows:
            window.hide()
        if self.window.project_window:
            self.window.project_window.hide()
        self.window.hide()
        if self.window.conn:
            self.window.conn.disconnect()
        QtWidgets.QApplication.processEvents()
        self.application.quit()
        QtWidgets.QApplication.processEvents()

    def test_changes(self):
        len_prev = self.proj_window.changes.count()
        # make a change
        self.window.waypoints_model.invert_direction()
        # save it with a comment
        self.window.save_wp_mscolab(comment="dummy save")
        QtWidgets.QApplication.processEvents()
        # test doesn't work without the sleep, because of delay in adding change and commit
        time.sleep(2)
        # change again for db consistency
        self.window.waypoints_model.invert_direction()
        self.window.save_wp_mscolab(comment="dummy save")
        QtWidgets.QApplication.processEvents()
        time.sleep(2)
        # fetch wp/chats/project
        self.window.reload_window(self.window.active_pid)
        QtWidgets.QApplication.processEvents()
        self.proj_window.load_all_changes()
        QtWidgets.QApplication.processEvents()
        len_after = self.proj_window.changes.count()
        # test change render
        assert len_prev == (len_after - 2)

    def test_undo(self):
        old_count = self.proj_window.changes.count()
        self._activate_change_at_index(0)
        QtWidgets.QApplication.processEvents()
        self.proj_window.request_undo_mscolab(self.proj_window.changes.currentIndex())
        # wait till server emits event to reload
        time.sleep(3)
        self.proj_window.load_all_changes()
        QtWidgets.QApplication.processEvents()
        assert self.proj_window.changes.count() == old_count + 1
        # delete the changes
        with self.app.app_context():
            Change.query.filter_by(comment="dummy save").delete()
            db.session.commit()

    def test_chat(self):
        self.proj_window.messageText.setPlainText('some - message')
        QtTest.QTest.mouseClick(self.proj_window.sendMessage, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        # delete message from db here
        # wait till server processes the change
        time.sleep(3)
        with self.app.app_context():
            assert Message.query.filter_by(text='some - message').count() == 1
            Message.query.filter_by(text='some - message').delete()
            db.session.commit()

    def test_users(self):
        assert self.proj_window.collaboratorsList.count() == 3
        # test remove/add/update collaborator
        self.proj_window.username.setText('c')
        QtTest.QTest.mouseClick(self.proj_window.delete_1, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        self.proj_window.accessLevel.setCurrentIndex(0)
        QtTest.QTest.mouseClick(self.proj_window.add, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()
        self.proj_window.accessLevel.setCurrentIndex(2)
        QtTest.QTest.mouseClick(self.proj_window.modify, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()

    def _connect_to_mscolab(self):
        self.window.url.setEditText("http://localhost:8084")
        QtTest.QTest.mouseClick(self.window.connectMscolab, QtCore.Qt.LeftButton)
        time.sleep(0.5)

    def _login(self):
        # login
        self._connect_to_mscolab()
        self.window.emailid.setText('a')
        self.window.password.setText('a')
        QtTest.QTest.mouseClick(self.window.loginButton, QtCore.Qt.LeftButton)
        QtWidgets.QApplication.processEvents()

    def _activate_project_at_index(self, index):
        item = self.window.listProjects.item(index)
        point = self.window.listProjects.visualItemRect(item).center()
        QtTest.QTest.mouseClick(self.window.listProjects.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.mouseDClick(self.window.listProjects.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()

    def _activate_change_at_index(self, index):
        item = self.proj_window.changes.item(index)
        point = self.proj_window.changes.visualItemRect(item).center()
        QtTest.QTest.mouseClick(self.proj_window.changes.viewport(), QtCore.Qt.LeftButton, pos=point)
        QtWidgets.QApplication.processEvents()
        QtTest.QTest.keyClick(self.proj_window.changes.viewport(), QtCore.Qt.Key_Return)
        QtWidgets.QApplication.processEvents()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_add_user.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_addUserDialog(object):
    def setupUi(self, addUserDialog):
        addUserDialog.setObjectName("addUserDialog")
        addUserDialog.resize(413, 275)
        self.buttonBox = QtWidgets.QDialogButtonBox(addUserDialog)
        self.buttonBox.setGeometry(QtCore.QRect(100, 230, 301, 32))
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.label = QtWidgets.QLabel(addUserDialog)
        self.label.setGeometry(QtCore.QRect(90, 80, 59, 15))
        self.label.setObjectName("label")
        self.label_2 = QtWidgets.QLabel(addUserDialog)
        self.label_2.setGeometry(QtCore.QRect(78, 110, 71, 20))
        self.label_2.setObjectName("label_2")
        self.label_3 = QtWidgets.QLabel(addUserDialog)
        self.label_3.setGeometry(QtCore.QRect(30, 140, 111, 20))
        self.label_3.setObjectName("label_3")
        self.label_4 = QtWidgets.QLabel(addUserDialog)
        self.label_4.setGeometry(QtCore.QRect(60, 170, 81, 20))
        self.label_4.setObjectName("label_4")
        self.emailid = QtWidgets.QLineEdit(addUserDialog)
        self.emailid.setGeometry(QtCore.QRect(150, 80, 211, 23))
        self.emailid.setObjectName("emailid")
        self.password = QtWidgets.QLineEdit(addUserDialog)
        self.password.setGeometry(QtCore.QRect(150, 110, 141, 23))
        self.password.setObjectName("password")
        self.rePassword = QtWidgets.QLineEdit(addUserDialog)
        self.rePassword.setGeometry(QtCore.QRect(150, 140, 141, 23))
        self.rePassword.setObjectName("rePassword")
        self.username = QtWidgets.QLineEdit(addUserDialog)
        self.username.setGeometry(QtCore.QRect(150, 170, 141, 23))
        self.username.setObjectName("username")

        self.retranslateUi(addUserDialog)
        self.buttonBox.accepted.connect(addUserDialog.accept)
        self.buttonBox.rejected.connect(addUserDialog.reject)
        QtCore.QMetaObject.connectSlotsByName(addUserDialog)

    def retranslateUi(self, addUserDialog):
        _translate = QtCore.QCoreApplication.translate
        addUserDialog.setWindowTitle(_translate("addUserDialog", "Add user"))
        self.label.setText(_translate("addUserDialog", "emailid"))
        self.label_2.setText(_translate("addUserDialog", "password"))
        self.label_3.setText(_translate("addUserDialog", "reenter password"))
        self.label_4.setText(_translate("addUserDialog", "username"))




# -*- coding: utf-8 -*-
"""

    mslib.msui.mscolab
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Window to display authentication and project details for mscolab


    To better understand of the code, look at the 'ships' example from
    chapter 14/16 of 'Rapid GUI Programming with Python and Qt: The
    Definitive Guide to PyQt Programming' (Mark Summerfield).

    This file is part of mss.

    :copyright: Copyright 2019- Shivashis Padhi
    :license: APACHE-2.0, see LICENSE for details.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from mslib.msui.mss_qt import QtGui, QtWidgets, QtCore, get_save_filename
from mslib.msui.mss_qt import ui_mscolab_window as ui
from mslib.msui.mss_qt import ui_add_user_dialog as add_user_ui
from mslib.msui.mss_qt import ui_add_project_dialog as add_project_ui
from mslib.msui import MissionSupportSystemDefaultConfig as mss_default
from mslib.msui.icons import icons
from mslib.msui import flighttrack as ft
from mslib.msui import topview, sideview, tableview
from mslib.msui import socket_control as sc
from mslib.msui import mscolab_project as mp
from mslib.utils import load_settings_qsettings, save_settings_qsettings
from mslib.utils import config_loader

import logging
import requests
import json
import fs

MSCOLAB_URL_LIST = QtGui.QStandardItemModel()


def add_mscolab_urls(combo_box, url_list):
    combo_box_urls = [combo_box.itemText(_i) for _i in range(combo_box.count())]
    for url in (_url for _url in url_list if _url not in combo_box_urls):
        combo_box.addItem(url)


class MSSMscolabWindow(QtWidgets.QMainWindow, ui.Ui_MSSMscolabWindow):
    """PyQt window implementing mscolab window
    """
    name = "Mscolab"
    identifier = None
    viewCloses = QtCore.pyqtSignal(name="viewCloses")

    def __init__(self, parent=None, data_dir=mss_default.mss_dir, mscolab_server_url=mss_default.mscolab_server_url):
        """Set up user interface
        """
        super(MSSMscolabWindow, self).__init__(parent)
        self.setupUi(self)
        self.loggedInWidget.hide()
        self.setWindowIcon(QtGui.QIcon(icons('64x64')))
        # if token is None, not authorized, else authorized
        self.token = None
        self.loginButton.clicked.connect(self.authorize)
        self.logoutButton.clicked.connect(self.logout)
        self.topview.clicked.connect(self.open_topview)
        self.sideview.clicked.connect(self.open_sideview)
        self.tableview.clicked.connect(self.open_tableview)
        self.save_ft.clicked.connect(self.save_wp_mscolab)
        self.fetch_ft.clicked.connect(self.reload_wps_from_server)
        self.autoSave.stateChanged.connect(self.autosave_emit)
        self.projWindow.clicked.connect(self.open_project_window)
        self.addProject.clicked.connect(self.add_project_handler)
        self.addUser.clicked.connect(self.add_user_handler)
        self.export_2.clicked.connect(self.handle_export)
        self.connectMscolab.clicked.connect(self.connect_handler)
        self.disconnectMscolab.clicked.connect(self.disconnect_handler)

        # int to store active pid
        self.active_pid = None
        # storing access_level to save network call
        self.access_level = None
        # store active_flight_path here as object
        self.waypoints_model = None
        # store a reference of window in class
        self.open_windows_mscolab = []
        # connection object to interact with sockets
        self.conn = None
        # to store tempfile name
        self.fname_temp = ""
        # store window instances
        self.active_windows = []
        # assign ids to view-window
        self.id_count = 0
        # project window
        self.project_window = None
        self.disable_action_buttons()
        # set data dir, uri
        self.data_dir = data_dir
        self.mscolab_server_url = None
        # autosave status
        self.autosave_status = None

        # disabling login, add user button. they are enabled when url is connected
        self.loginButton.setEnabled(False)
        self.addUser.setEnabled(False)
        self.disconnectMscolab.setEnabled(False)

        self.url.setEditable(True)
        self.url.setModel(MSCOLAB_URL_LIST)
        # fill value of mscolab url from config
        default_MSCOLAB = config_loader(dataset="default_MSCOLAB", default=mss_default.default_MSCOLAB)
        add_mscolab_urls(self.url, default_MSCOLAB)
        # fill value of mscolab url if found in QSettings storage
        self.settings = load_settings_qsettings('mscolab', default_settings={'mscolab_url': None, 'auth': None})
        if self.settings['mscolab_url'] is not None:
            add_mscolab_urls(self.url, [self.settings['mscolab_url']])

    def disconnect_handler(self):
        self.logout()
        # enable and disable right buttons
        self.disconnectMscolab.setEnabled(False)
        self.loginButton.setEnabled(False)
        self.addUser.setEnabled(False)
        self.connectMscolab.setEnabled(True)
        # set mscolab_server_url to None
        self.mscolab_server_url = None

    def show_info(self, text):
        self.error_dialog = QtWidgets.QErrorMessage()
        self.error_dialog.showMessage(text)

    def connect_handler(self):
        try:
            url = str(self.url.currentText())
            r = requests.get(url)
            if r.text == "Mscolab server":
                # assign new url to self.mscolab_server_url
                self.mscolab_server_url = url
                self.status.setText("Status: connected")
                # enable and disable right buttons
                self.loginButton.setEnabled(True)
                self.addUser.setEnabled(True)
                self.disconnectMscolab.setEnabled(True)
                self.connectMscolab.setEnabled(False)
                self.settings["mscolab_url"] = url
                save_settings_qsettings('mscolab', self.settings)
                return
        except requests.exceptions.ConnectionError:
            logging.debug("mscolab server isn't active")
        except requests.exceptions.InvalidSchema:
            logging.debug("invalid schema of url")
        except requests.exceptions.InvalidURL:
            logging.debug("invalid url")
        except Exception as e:
            logging.debug("Error {}".format(str(e)))
        # inform user that url is invalid
        self.show_info("Invalid url, please try again!")

    def handle_export(self):
        # ToDo when autosave mode gets upgraded, have to fetch from remote
        file_path = get_save_filename(
            self, "Save fight track", "", "Flight Track Files (*.ftml)")
        if file_path is not None:
            f_name = fs.path.basename(file_path)
            f_dir = fs.open_fs(fs.path.dirname(file_path))
            temp_name = 'tempfile_mscolab.ftml'
            temp_dir = fs.open_fs(self.data_dir)
            fs.copy.copy_file(temp_dir, temp_name, f_dir, f_name)

    def disable_action_buttons(self):
        # disable some buttons to be activated after successful login or project activate
        self.addProject.setEnabled(False)
        self.save_ft.setEnabled(False)
        self.fetch_ft.setEnabled(False)
        self.topview.setEnabled(False)
        self.sideview.setEnabled(False)
        self.tableview.setEnabled(False)
        self.projWindow.setEnabled(False)
        self.autoSave.setEnabled(False)
        self.export_2.setEnabled(False)

    def add_project_handler(self):
        if self.token is None:
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('Please login to use this feature')
            return
        else:
            logging.debug(self.token)
        self.proj_diag = QtWidgets.QDialog()
        self.add_proj_dialog = add_project_ui.Ui_addProjectDialog()
        self.add_proj_dialog.setupUi(self.proj_diag)
        self.add_proj_dialog.f_content = None
        self.add_proj_dialog.buttonBox.accepted.connect(self.add_project)
        # enable accepted only if path and description are not none
        self.add_proj_dialog.buttonBox.setEnabled(False)
        self.add_proj_dialog.path.textChanged.connect(self.check_and_enable_project_accept)
        self.add_proj_dialog.description.textChanged.connect(self.check_and_enable_project_accept)
        self.add_proj_dialog.browse.clicked.connect(self.set_exported_file)
        self.proj_diag.show()

    def check_and_enable_project_accept(self):
        if(self.add_proj_dialog.path.text() != "" and self.add_proj_dialog.description.toPlainText() != ""):
            self.add_proj_dialog.buttonBox.setEnabled(True)

    def set_exported_file(self):
        file_path = QtWidgets.QFileDialog.getOpenFileName()[0]
        if file_path == "":
            return
        f_name = fs.path.basename(file_path)
        f_dir = fs.open_fs(fs.path.dirname(file_path))
        f_content = f_dir.readtext(f_name)
        self.add_proj_dialog.f_content = f_content
        self.add_proj_dialog.selectedFile.setText(f_name)

    def add_project(self):
        path = self.add_proj_dialog.path.text()
        description = self.add_proj_dialog.description.toPlainText()
        # ToDo if path and description is null alert user
        if not path:
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('Path can\'t be empty')
            return
        elif not description:
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('Description can\'t be empty')
            return

        data = {
            "token": self.token,
            "path": path,
            "description": description
        }
        if self.add_proj_dialog.f_content is not None:
            data["content"] = self.add_proj_dialog.f_content
        r = requests.post('{}/create_project'.format(self.mscolab_server_url), data=data)
        if r.text == "True":
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('Your project was created successfully')
            self.add_projects()
            p_id = self.get_recent_pid()
            self.conn.handle_new_room(p_id)
        else:
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('The path already exists')

    def add_user_handler(self):
        self.user_diag = QtWidgets.QDialog()
        self.add_user_dialog = add_user_ui.Ui_addUserDialog()
        self.add_user_dialog.setupUi(self.user_diag)
        self.add_user_dialog.buttonBox.accepted.connect(self.add_user)
        self.user_diag.show()

    def add_user(self):
        emailid = self.add_user_dialog.emailid.text()
        password = self.add_user_dialog.password.text()
        re_password = self.add_user_dialog.rePassword.text()
        username = self.add_user_dialog.username.text()
        if password == re_password:
            data = {
                "email": emailid,
                "password": password,
                "username": username
            }
            r = requests.post('{}/register'.format(self.mscolab_server_url), data=data)
            if r.text == "True":
                self.error_dialog = QtWidgets.QErrorMessage()
                self.error_dialog.showMessage('You are registered, you can now log in.')
            else:
                self.error_dialog = QtWidgets.QErrorMessage()
                self.error_dialog.showMessage('Oh no, your emailid is either invalid or taken')
        else:
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('Oh no, your passwords don\'t match')

    def open_project_window(self):
        if self.active_pid is None:
            return
        if self.project_window is not None:
            return
        view_window = mp.MSColabProjectWindow(self.token, self.active_pid, self.conn,
                                              self.access_level, parent=self.projWindow,
                                              mscolab_server_url=self.mscolab_server_url)
        view_window.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        view_window.viewCloses.connect(self.close_project_window)
        view_window.reloadWindows.connect(self.reload_windows_slot)
        self.project_window = view_window
        self.project_window.show()

    def close_project_window(self):
        self.project_window = None

    def autosave_emit(self):
        # emit signal to server to enable or disable
        if self.active_pid is None:
            return

        # ToDo replace checkbox by text if normal user
        # ToDo save file as backup before loading what's in admin

        logging.debug(self.autoSave.isChecked())
        if self.autoSave.isChecked():
            self.conn.emit_autosave(self.token, self.active_pid, 1)
        else:
            self.conn.emit_autosave(self.token, self.active_pid, 0)

    def authorize(self):
        emailid = self.emailid.text()
        password = self.password.text()
        # to prevent someone nearby from seeing the id, password
        self.emailid.setText('')
        self.password.setText('')
        data = {
            "email": emailid,
            "password": password
        }
        r = requests.post(self.mscolab_server_url + '/token', data=data)
        if r.text == "False":
            # popup that has wrong credentials
            self.error_dialog = QtWidgets.QErrorMessage()
            self.error_dialog.showMessage('Oh no, your credentials were incorrect.')
            pass
        else:
            # remove the login modal and put text there
            _json = json.loads(r.text)
            self.token = _json["token"]
            self.user = _json["user"]
            self.label.setText("logged in as: " + _json["user"]["username"])
            self.loggedInWidget.show()
            self.loginWidget.hide()

            self.add_projects()

            # create socket connection here
            self.conn = sc.ConnectionManager(self.token, user=self.user, mscolab_server_url=self.mscolab_server_url)
            self.conn.signal_reload.connect(self.reload_window)
            self.conn.signal_autosave.connect(self.autosave_toggle)
            self.conn.signal_new_permission.connect(self.render_new_permission)
            self.conn.signal_update_permission.connect(self.handle_update_permission)
            self.conn.signal_revoke_permission.connect(self.handle_revoke_permission)
            # activate add project button here
            self.addProject.setEnabled(True)

    def add_projects(self):
        # add projects
        data = {
            "token": self.token
        }
        r = requests.get(self.mscolab_server_url + '/projects', data=data)
        _json = json.loads(r.text)
        projects = _json["projects"]
        self.add_projects_to_ui(projects)

    def get_recent_pid(self):
        """
        get most recent project's p_id
        # ToDo can be merged with get_recent_project
        """
        data = {
            "token": self.token
        }
        r = requests.get(self.mscolab_server_url + '/projects', data=data)
        _json = json.loads(r.text)
        projects = _json["projects"]
        return projects[-1]["p_id"]

    def get_recent_project(self):
        """
        get most recent project
        """
        data = {
            "token": self.token
        }
        r = requests.get(self.mscolab_server_url + '/projects', data=data)
        _json = json.loads(r.text)
        projects = _json["projects"]
        return projects[-1]

    def add_projects_to_ui(self, projects):
        logging.debug("adding projects to ui")
        self.listProjects.clear()
        selectedProject = None
        for project in projects:
            project_desc = '{} - {}'.format(project['path'], project["access_level"])
            widgetItem = QtWidgets.QListWidgetItem(project_desc, parent=self.listProjects)
            widgetItem.p_id = project["p_id"]
            widgetItem.access_level = project["access_level"]
            if widgetItem.p_id == self.active_pid:
                selectedProject = widgetItem
            self.listProjects.addItem(widgetItem)
        if selectedProject is not None:
            self.listProjects.setCurrentItem(selectedProject)
            self.listProjects.itemActivated.emit(selectedProject)
        self.listProjects.itemActivated.connect(self.set_active_pid)

    def set_active_pid(self, item):
        # remove all windows if the current active_pid is not selected p_id
        if item.p_id != self.active_pid:
            # close all hanging window
            for window in self.active_windows:
                window.close()
            # show autosave button, and empty autosaveStatus
            self.autoSave.setVisible(True)
            self.autosaveStatus.setText("")
        # set active_pid here
        self.active_pid = item.p_id
        self.access_level = item.access_level
        # set active flightpath here
        self.load_wps_from_server()

        # enable project specific buttons
        self.save_ft.setEnabled(True)
        self.fetch_ft.setEnabled(True)
        self.topview.setEnabled(True)
        self.sideview.setEnabled(True)
        self.tableview.setEnabled(True)
        self.projWindow.setEnabled(True)
        self.autoSave.setEnabled(True)
        self.export_2.setEnabled(True)
        # configuring autosave button
        data = {
            "token": self.token,
            "p_id": self.active_pid
        }
        r = requests.get(self.mscolab_server_url + '/project_details', data=data)
        _json = json.loads(r.text)

        if _json["autosave"] is True:
            # one time activate
            self.autoSave.blockSignals(True)
            self.autoSave.setChecked(True)
            self.autoSave.blockSignals(False)
            self.save_ft.setEnabled(False)
            self.fetch_ft.setEnabled(False)
            # connect data change to handler
            self.waypoints_model.dataChanged.connect(self.handle_data_change)
            # enable autosave
            self.autosave_status = True
        else:
            self.autoSave.blockSignals(True)
            self.autoSave.setChecked(False)
            self.autoSave.blockSignals(False)
            self.autosave_status = False

        # hide autosave button if access_level is non-admin
        if self.access_level == "viewer" or self.access_level == "collaborator":
            self.autoSave.setVisible(False)
            # set autosave status
            if _json["autosave"]:
                self.autosaveStatus.setText("Autosave is enabled")
            else:
                self.autosaveStatus.setText("Autosave is disabled")
        else:
            self.autosaveStatus.setText("")
            self.autoSave.setVisible(True)

        # change font style for selected
        font = QtGui.QFont()
        for i in range(self.listProjects.count()):
            self.listProjects.item(i).setFont(font)
        font.setBold(True)
        item.setFont(font)

    def reload_wps_from_server(self):
        if self.active_pid is None:
            return
        self.load_wps_from_server()
        for window in self.active_windows:
            # set active flight track
            window.setFlightTrackModel(self.waypoints_model)
            # redraw figure *only for canvas based window, not tableview*
            if hasattr(window, 'mpl'):
                window.mpl.canvas.waypoints_interactor.redraw_figure()

    def load_wps_from_server(self):
        data = {
            "token": self.token,
            "p_id": self.active_pid
        }
        r = requests.get(self.mscolab_server_url + '/get_project', data=data)
        ftml = json.loads(r.text)["content"]
        data_dir = fs.open_fs(self.data_dir)
        data_dir.writetext('tempfile_mscolab.ftml', ftml)
        fname_temp = fs.path.combine(self.data_dir, 'tempfile_mscolab.ftml')
        self.fname_temp = fname_temp
        self.waypoints_model = ft.WaypointsTableModel(filename=fname_temp)

        # connect change events viewwindow HERE to emit file-save
        self.waypoints_model.dataChanged.connect(self.handle_data_change)

    def open_topview(self):
        # showing dummy info dialog
        if self.active_pid is None:
            return
        self.create_view_window("topview")

    def open_sideview(self):
        # showing dummy info dialog
        if self.active_pid is None:
            return
        self.create_view_window("sideview")

    def open_tableview(self):
        # showing dummy info dialog
        if self.active_pid is None:
            return
        self.create_view_window("tableview")

    def create_view_window(self, _type):
        view_window = None
        if _type == "topview":
            view_window = topview.MSSTopViewWindow(model=self.waypoints_model,
                                                   parent=self.listProjects,
                                                   _id=self.id_count)
            view_window.view_type = "topview"
        elif _type == "sideview":
            view_window = sideview.MSSSideViewWindow(model=self.waypoints_model,
                                                     parent=self.listProjects,
                                                     _id=self.id_count)
            view_window.view_type = "sideview"
        else:
            view_window = tableview.MSSTableViewWindow(model=self.waypoints_model,
                                                       parent=self.listProjects,
                                                       _id=self.id_count)
            view_window.view_type = "tableview"
        if self.access_level == "viewer":
            self.disable_navbar_action_buttons(_type, view_window)

        view_window.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        view_window.show()
        view_window.viewClosesId.connect(self.handle_view_close)
        self.active_windows.append(view_window)

        # increment id_count
        self.id_count += 1

    def disable_navbar_action_buttons(self, _type, view_window):
        """
        _type: view type (topview, sideview, tableview)
        view_window: PyQt view window

        function disables some control, used if access_level is not appropriate
        """
        if _type == "topview" or _type == "sideview":
            actions = view_window.mpl.navbar.actions()
            for action in actions:
                action_text = action.text()
                if action_text == "Ins WP" or action_text == "Del WP" or action_text == "Mv WP":
                    action.setEnabled(False)
        else:
            # _type == tableview
            view_window.btAddWayPointToFlightTrack.setEnabled(False)
            view_window.btCloneWaypoint.setEnabled(False)
            view_window.btDeleteWayPoint.setEnabled(False)
            view_window.btInvertDirection.setEnabled(False)

    def enable_navbar_action_buttons(self, _type, view_window):
        """
        _type: view type (topview, sideview, tableview)
        view_window: PyQt view window

        function enables some control, used if access_level is appropriate
        """
        if _type == "topview" or _type == "sideview":
            actions = view_window.mpl.navbar.actions()
            for action in actions:
                action_text = action.text()
                if action_text == "Ins WP" or action_text == "Del WP" or action_text == "Mv WP":
                    action.setEnabled(True)
        else:
            # _type == tableview
            view_window.btAddWayPointToFlightTrack.setEnabled(True)
            view_window.btCloneWaypoint.setEnabled(True)
            view_window.btDeleteWayPoint.setEnabled(True)
            view_window.btInvertDirection.setEnabled(True)

    def logout(self):
        # check if autosave is enabled
        # ToDo for non-admins who will get autosave hidden
        self.w = QtWidgets.QWidget()
        if (not self.autoSave.isChecked()) and (self.active_pid is not None):
            qm = QtWidgets.QMessageBox
            reply = qm.question(self.w, 'Continue?',
                                'Autosave is disabled, save your changes locally before continuing!',
                                qm.Yes,
                                qm.No)
            if reply == QtWidgets.QMessageBox.No:
                return
        # delete token and show login widget-items
        self.token = None
        # delete active-project-id
        self.active_pid = None
        # delete active access_level
        self.access_level = None
        # clear projects list here
        self.loggedInWidget.hide()
        self.loginWidget.show()
        # clear project listing
        self.listProjects.clear()
        # disconnect socket
        if self.conn is not None:
            self.conn.disconnect()
            self.conn = None
        # close all hanging window
        for window in self.active_windows:
            window.close()
        # close project window if active
        if self.project_window is not None:
            self.project_window.close()
        # show autosave button, and empty autosaveStatus
        self.autoSave.setVisible(True)
        self.autosaveStatus.setText("")

        self.disable_action_buttons()

    def save_wp_mscolab(self, comment=None):
        if self.active_pid is not None:
            # to save to temp file
            xml_text = self.waypoints_model.save_to_mscolab()
            # to emit to mscolab
            self.conn.save_file(self.token, self.active_pid, xml_text, comment=comment)

    @QtCore.Slot(int, int, str)
    def handle_update_permission(self, p_id, u_id, access_level):
        """
        p_id: project id
        u_id: user id
        access_level: updated access level

        function updates existing permissions and related control availability
        """
        if u_id == self.user["id"]:
            # update table of projects
            for i in range(self.listProjects.count()):
                item = self.listProjects.item(i)
                if item.p_id == p_id:
                    desc = item.text().split('-')
                    desc[-1] = access_level
                    desc = '-'.join(desc)
                    item.setText(desc)
                    item.p_id = p_id
                    item.access_level = access_level
            if p_id != self.active_pid:
                return
            self.access_level = access_level
            # update project window's control if open
            if self.project_window is not None:
                self.project_window.check_permission_and_render_control(self.access_level)
            # update view window nav elements if open
            for window in self.active_windows:
                _type = window.view_type
                if self.access_level == "viewer":
                    self.disable_navbar_action_buttons(_type, window)
                else:
                    self.enable_navbar_action_buttons(_type, window)
            # update autosave stats
            if self.access_level == "admin" or self.access_level == "creator":
                # enable autosave set it to checked status
                self.autoSave.setVisible(True)
                if self.autosave_status is True:
                    self.autoSave.blockSignals(True)
                    self.autoSave.setChecked(True)
                    self.autoSave.blockSignals(False)
                else:
                    self.autoSave.blockSignals(True)
                    self.autoSave.setChecked(False)
                    self.autoSave.blockSignals(False)
                self.autosaveStatus.setText("")
            else:
                # disable autosave, set status text
                self.autoSave.setVisible(False)
                if self.autosave_status is True:
                    self.autosaveStatus.setText("Autosave is enabled!")
                else:
                    self.autosaveStatus.setText("Autosave is enabled!")

        # update project window if open
        if self.project_window is not None:
            self.project_window.load_users()

    @QtCore.Slot(int, int)
    def handle_revoke_permission(self, p_id, u_id):
        if u_id == self.user["id"]:
            # Check if the user has opened any windows of revoked project and close them
            if self.active_pid == p_id:
                for window in self.active_windows:
                    window.close()
                if self.project_window is not None:
                    self.project_window.close()
                self.active_pid = None

            # Update project list
            remove_item = None
            for i in range(self.listProjects.count()):
                item = self.listProjects.item(i)
                if item.p_id == p_id:
                    remove_item = item
            if remove_item is not None:
                self.listProjects.takeItem(self.listProjects.row(remove_item))

    @QtCore.Slot()
    def reload_windows_slot(self):
        self.reload_window(self.active_pid)

    @QtCore.Slot(int, int)
    def render_new_permission(self, p_id, u_id):
        """
        p_id: project id
        u_id: user id

        to render new permission if added
        """
        data = {
            'token': self.token
        }
        r = requests.get(self.mscolab_server_url + '/user', data=data)
        _json = json.loads(r.text)
        if _json['user']['id'] == u_id:
            project = self.get_recent_project()
            project_desc = '{} - {}'.format(project['path'], project["access_level"])
            widgetItem = QtWidgets.QListWidgetItem(project_desc, parent=self.listProjects)
            widgetItem.p_id = project["p_id"]
            widgetItem.access_level = project["access_level"]
            self.listProjects.addItem(widgetItem)
        if self.project_window is not None:
            self.project_window.load_users()

    @QtCore.Slot(int)
    def reload_window(self, value):
        if self.active_pid != value or self.autosave_status is False:
            return
        logging.debug("reloading window")
        # ask the user in dialog if he wants the change, only for autosave mode
        # toDo preview of the change
        self.reload_wps_from_server()
        # reload changes in project window if it exists
        if self.project_window is not None:
            self.project_window.load_all_changes()
            self.project_window.load_all_messages()

    @QtCore.Slot(int)
    def handle_view_close(self, value):
        logging.debug("removing stale window")
        for index, window in enumerate(self.active_windows):
            if window._id == value:
                del self.active_windows[index]

    @QtCore.Slot(QtCore.QModelIndex, QtCore.QModelIndex)
    def handle_data_change(self, index1, index2):
        # if autosave isn't checked, don't save. (in future, this might be hidden # ToDo)
        if self.autoSave.isChecked():
            self.save_wp_mscolab()

    @QtCore.Slot(int, int)
    def autosave_toggle(self, enable, p_id):
        # return if it's for a different process
        if p_id != self.active_pid:
            return
        if enable:
            # enable autosave, disable save button
            self.autosave_status = True
            self.save_ft.setEnabled(False)
            self.fetch_ft.setEnabled(False)
            # reload window
            self.reload_wps_from_server()
            self.waypoints_model.dataChanged.connect(self.handle_data_change)
            self.autosaveStatus.setText("Autosave is enabled")

        else:
            self.autosave_status = False
            # disable autosave, enable save button
            self.save_ft.setEnabled(True)
            self.fetch_ft.setEnabled(True)
            # connect change events viewwindow HERE to emit file-save
            # ToDo - remove hack to disconnect this handler
            self.waypoints_model.dataChanged.disconnect(self.handle_data_change)
            self.autosaveStatus.setText("Autosave is disabled")

    def setIdentifier(self, identifier):
        self.identifier = identifier

    def closeEvent(self, event):
        if self.conn:
            self.conn.disconnect()
